#include "Camera.h"
#include "Utils.h"

//#define CAM_DEBUG(x) { PRINTOUT("CAMDEBUG " << x); printDebugInfo(); }
#define CAM_DEBUG(x)

const float Camera::THETA_CLAMP = glm::radians(89.9f);

glm::mat4 Camera::getViewMatrix() {

	// to move the camera from (0,0) to its current position/orientation:
	// Translate to (0,0,r), rotate around x (theta), rotate around y (phi), translate to focus: 
	// T(focus)*Ry(-phi)*Rx(-theta)*T(0,0,r)
	//
	// for the view matrix we need the inverse operation:
	// (A*B*C*D)^-1 = (D^-1)*(C^-1)*(B^-1)*(A^-1)
	// T(0,0,-r)*Rx(theta)*Ry(phi)*T(-focus)

	// This works, but calcs the inverse exlicitly:
	//glm::mat4 mat = glm::mat4(1.0f);
	//mat = glm::translate(mat, m_offset);					// 4 move to focus
	//mat = glm::rotate(mat, -m_phi, glm::vec3(0, 1, 0));	// 3 rot y
	//mat = glm::rotate(mat, -m_theta, glm::vec3(1, 0, 0));	// 2 rot x
	//mat = glm::translate(mat, glm::vec3(0, 0, m_r));		// 1 move to view distance
	//mat = glm::inverse(mat);								// invert

	// without having to calc an inverse (above steps separately inverted in rev. order):
	glm::mat4 mat = glm::mat4(1.0f);
	mat = glm::translate(mat, glm::vec3(0, 0, -m_r));		// 1 move to view distance (inverse)
	mat = glm::rotate(mat, m_theta, glm::vec3(1, 0, 0));	// 2 rot x (inverse)
	mat = glm::rotate(mat, m_phi, glm::vec3(0, 1, 0));		// 3 rot y (inverse)
	mat = glm::translate(mat, -m_offset);					// 4 move to focus (inverse)

	return mat;
}

void Camera::reset() {
	m_phi    = m_reset_phi;
	m_theta  = m_reset_theta;
	m_r      = m_reset_r;
	m_offset = m_reset_offset;
	m_min_rad = m_reset_min_rad;
	m_max_rad = m_reset_max_rad;
	CAM_DEBUG("reset")
}

void Camera::moveTheta(float delta) {
	m_theta += delta;
	if (m_theta > THETA_CLAMP) {
		m_theta = THETA_CLAMP;
	} else if (m_theta < -THETA_CLAMP) {
		m_theta = -THETA_CLAMP;
	}
}

void Camera::movePhi(float delta) {
	m_phi += delta;
}

void Camera::moveRad(float delta) {
	m_r += delta;
	if (m_r < m_min_rad) m_r = m_min_rad;
	if (m_r > m_max_rad) m_r = m_max_rad;
}

void Camera::moveRightUp(float deltaRight, float deltaUp) {
	// basic camera transform
	glm::mat4 mat = glm::mat4(1.0f);
	mat = glm::rotate(mat, -m_phi, glm::vec3(0, 1, 0));		// 3 rot y
	mat = glm::rotate(mat, -m_theta, glm::vec3(1, 0, 0));	// 2 rot x
	mat = glm::translate(mat, glm::vec3(0, 0, m_r));		// 1 move to view distance

	// claculate new right and up vector by transforming
	// the x and y unit vectors
	glm::vec3 right = mat * glm::vec4(1, 0, 0, 0); // new right; don't care about translation, w can be 0
	glm::vec3 up    = mat * glm::vec4(0, 1, 0, 0); // new up;    don't care about translation, w can be 0

	m_offset += deltaRight * right + deltaUp * up;
}

void Camera::setFocus(glm::vec3 focus) {
	m_offset = focus;
}

glm::mat4 Camera::getWorldTransform() {
	// return a transformation matrix to move the world coordinate system to the camera coordinate system

	// basic camera transform
	glm::mat4 mat = glm::mat4(1.0f);
	mat = glm::translate(mat, m_offset);					// 4 move to focus
	mat = glm::rotate(mat, -m_phi, glm::vec3(0, 1, 0));		// 3 rot y
	mat = glm::rotate(mat, -m_theta, glm::vec3(1, 0, 0));	// 2 rot x
	mat = glm::translate(mat, glm::vec3(0, 0, m_r));		// 1 move to view distance
	return mat;
}

glm::vec3 Camera::getWorldPosition() {
	// return the camera position in world space coordinates
	glm::vec4 v = getWorldTransform() * glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
	return glm::vec3(v.x, v.y, v.z);
}

void Camera::getOrientation(glm::vec3 *forward, glm::vec3 *right, glm::vec3 *up) {
	glm::mat4 wt = getWorldTransform();
	if (forward) *forward = glm::vec3(wt * glm::vec4(0.0f, 0.0f, -1.0f, 0.0f));
	if (right)   *right   = glm::vec3(wt * glm::vec4(1.0f, 0.0f, 0.0f, 0.0f));
	if (up)      *up      = glm::vec3(wt * glm::vec4(0.0f, 1.0f, 0.0f, 0.0f));
}

//glm::quat Camera::getOrientationQuaternion()
//{
//	//return glm::quat(glm::mat4(glm::mat3(getWorldTransform()))); // just to verify
//	return glm::quat(glm::vec3(-m_theta, -m_phi, 0.0f));           // it's the same
//}
//
//glm::quat Camera::getViewQuaternion()
//{
//	return glm::quat(glm::vec3(m_theta, m_phi, 0.0f));
//}

//void Camera::printDebugInfo() {
//	glm::vec3 f, r, u; getOrientation(&f, &r, &u);
//	
//	std::cout << "INFO: Cam world pos: " << DBG_VEC3(getWorldPosition()) << std::endl;
//	std::cout << "INFO: Cam params (phi,theta,r,focus): " << glm::degrees(m_phi) << " deg, " << glm::degrees(m_theta) << " deg, " << m_r << " " << DBG_VEC3(m_offset) << " min_rad " << m_min_rad << " max_rad " << m_max_rad << std::endl;
//	std::cout << "INFO: Cam reset  (phi,theta,r,focus): " << glm::degrees(m_reset_phi) << " deg, " << glm::degrees(m_reset_theta) << " deg, " << m_reset_r << " " << DBG_VEC3(m_reset_offset) << " min_rad " << m_reset_min_rad << " max_rad " << m_reset_max_rad << std::endl;
//	std::cout << "INFO: Cam forward " << DBG_VEC3(f) << ", right " << DBG_VEC3(r) << ", up " << DBG_VEC3(u) << std::endl;
//	std::cout << "INFO: Cam orientation quat " << DBG_QUAT(getOrientationQuaternion()) << std::endl;
//	std::cout << "INFO: Cam view        quat " << DBG_QUAT(getViewQuaternion()) << std::endl;
//}

void Camera::setParameters(float phi, float theta, float r, glm::vec3 offset) {
	m_phi = phi;
	m_theta = theta;
	m_r = r;
	m_offset = offset;
	CAM_DEBUG("setParameters")
}

void Camera::setTheta(float theta) {
	m_theta = theta;
}

void Camera::storeResetParameters()
{
	m_reset_phi    = m_phi;
	m_reset_theta  = m_theta;
	m_reset_r      = m_r;
	m_reset_offset = m_offset;
	m_reset_min_rad = m_min_rad;
	m_reset_max_rad = m_max_rad;
	CAM_DEBUG("storeResetParameters")
}
