#pragma once

#include <glm\gtc\matrix_transform.hpp>
#include <glm\glm.hpp>

/// Represents an arc-ball camera looking at the scene.
///
/// The current position and orientation is defined by four parameters:
/// - Theta:       Rotation around the X-axis
/// - Phi:         Rotation around the Y-axis
/// - Radius:      Distance from focal point
/// - Offset:      Coordinates of the focal point the camera is looking at
///
/// @note Theta-movement is clamped to &plusmn; 90 degrees to avoid upside-down viewing angles. This can be bypassed with setTheta().
class Camera {
public:
	static const float THETA_CLAMP; // = glm::radians(89.9f);

	Camera() { reset(); }

	inline float getPhi()   { return m_phi; }			///< Get the Phi angle.   @returns Phi
	inline float getTheta() { return m_theta; }			///< Get the Theta angle. @returns Theta
	inline float getRad()   { return m_r; }				///< Get the radius.      @returns Radius
	inline glm::vec3 getOffset() { return m_offset; }	///< Get the offset.      @returns Offset

	void moveTheta(float delta);	///< Adjust the Theta angle. @param delta Change to Theta
	void movePhi(float delta);		///< Adjust the Phi angle.   @param delta Change to Phi
	void moveRad(float delta);		///< Adjust the radius.      @param delta Change to Radius
	/// Move the camera offset.
	/// @param deltaRight Right(+)/left(-) movement
	/// @param deltaUp Up(+)/down(-) movement
	/// @note Up/down/left/right movement is relative to the current camera orientation.
	void moveRightUp(float deltaRight, float deltaUp);
	void reset();	///< Reset parameters to saved position.

	void setFocus(glm::vec3 focus);	///< Set the offset (focus) point. @param focus New offset point

	/// Set all camera position/orientation parameters at once.
	/// @param phi Phi angle
	/// @param theta Theta angle
	/// @param r Radius
	/// @param offset Offset (focus)
	void setParameters(float phi, float theta, float r, glm::vec3 offset);

	/// Set Theta angle manually.
	/// @attention This bypasses clamping of Theta!
	void setTheta(float theta); // unchecked!
	void storeResetParameters(); ///< Store current parameters.

	//void printDebugInfo();

	glm::mat4 getViewMatrix();		///< Get the view matrix of the camera. @returns View matrix
	glm::vec3 getWorldPosition();	///< Get the world-position of the camera. @returns World position

	/// Get the local coordinate system of the camera.
	///
	/// The returned vectors are in world-space.
	/// @param[out] forward Forward-direction
	/// @param[out] right   Right-direction
	/// @param[out] up      Up-direction
	void getOrientation(glm::vec3 *forward, glm::vec3 *right, glm::vec3 *up);
	//glm::quat getViewQuaternion();

	inline void  setMaxRad(float m) { m_max_rad = m; }	///< Set the maximum radius. @param m Maximum radius
	inline float getMaxRad() { return m_max_rad; }		///< Get the maximum radius. @returns Maximum radius
	inline void  setRad(float r) { m_r = r; moveRad(0.0f); /* clamp */ }	///< Set the radius. @param r radius

	// store projection params
	float m_fov;      // in radians
	float m_aspect;
	float m_near;
	float m_far;
private:
	glm::mat4 getWorldTransform();
	//glm::quat getOrientationQuaternion();

	// position (spherical coordinates) RELATIVE to focus-point
	float m_phi;
	float m_theta;
	float m_r;
	glm::vec3 m_offset; // camera offset (and focus-point)

	float m_reset_phi   = 0.0f;
	float m_reset_theta = 0.0f;
	float m_reset_r     = 6.0f;
	float m_reset_min_rad = 0.1f;
	float m_reset_max_rad = 500.0f;
	glm::vec3 m_reset_offset = glm::vec3(0.0f); // camera offset (and focus-point)

	float m_min_rad = 0.1f;
	float m_max_rad = 500.0f;
};
