#pragma once

#include "Utils.h"

/// Represents an OpenGL-Framebuffer.
/// @see function init() for details.
class Framebuffer {
public:
	static const int MAX_COLOR_ATTACHMENTS = 8;
	enum AttachmentType { 
		atNone,		///< no attachment
		atBuffer,	///< attachment as render buffer
		atTexture	///< attachment as texture
	};

	/// Creates an uninitialised Framebuffer
	inline Framebuffer() {};

	/// Create and initialise framebuffer.
	///
	/// Using this constructor is the same as using the plain constructor Framebuffer() followed by calling init().
	inline Framebuffer(int width, int height, AttachmentType colorAttachmentType, AttachmentType depthAttachmentType, int multisample = 0, int numColorAttachments = 1) {
		init(width, height, colorAttachmentType, depthAttachmentType, multisample, numColorAttachments);
	}
	inline ~Framebuffer() {
		glDeleteFramebuffers(1, &m_fbo);
		if (m_colorRB) glDeleteRenderbuffers(1, &m_colorRB);
		if (m_depthRB && externalDepthRenderBuffer == 0) glDeleteRenderbuffers(1, &m_depthRB);
		for (int i = 0; i < MAX_COLOR_ATTACHMENTS; i++)
			if (m_texColor[i] && externalColorTexture[i] == 0) glDeleteTextures(1, &(m_texColor[i]));
		if (m_texDepth) glDeleteTextures(1, &m_texDepth);
	}

	inline GLuint getFBO() { return m_fbo; }									///< Get the OpenGL framebuffer object (FBO). @returns FBO
	inline GLuint getColorTexture(int slot = 0) { return m_texColor[slot]; }	///< Get colour texture attachment handle. @param slot attachment number @returns texture handle
	inline GLuint getDepthTexture() { return m_texDepth; }						///< Get the depth buffer texture attachment handle. @returns texture handle
	inline GLuint getDepthRenderBuffer() { return m_depthRB; }					///< Get the depth buffer render buffer handle. @returns buffer handle

	/// Initialize framebuffer.
	///
	/// Colour and/or depth attachments can be created as textures or render buffers, or omitted entirely.
	/// It is also possible to specify already existing colour textures or an existing depth render buffer.
	///
	/// @param width               Width  of the viewport and textures/buffers
	/// @param height              Height of the viewport and textures/buffers
	/// @param colorAttachmentType Type of the colour attachment(s)
	/// @param depthAttachmentType Type of the depth attachment
	/// @param multisample         Number of multisamples (0 = no multisampling)
	/// @param numColorAttachments Number of colour attachments
	///
	/// Colour attachments will be created in the format specified by #COLOR_INT_FORMAT.
	/// The depth attachment will be created in the format specified by #DEPTH_INT_FORMAT.
	///
	/// To use an existing colour texture, set #externalColorTexture prior to calling init().
	/// Likewise, to use an existing depth render buffer, set #externalDepthRenderBuffer before calling init().
	///
	/// Textures can also be created as array-textures; see #colorTextureArraySize and #externalDepthRenderBuffer.
	inline void init(int width, int height, AttachmentType colorAttachmentType, AttachmentType depthAttachmentType, int multisample = 0, int numColorAttachments = 1) {
		if (m_fbo) FATAL_ERROR("Framebuffer: already inited");
		if (colorAttachmentType == atNone) numColorAttachments = 0;
		if (numColorAttachments > 1 && colorAttachmentType != atTexture) FATAL_ERROR("Framebuffer: Multiple color attachments only supported for textures");

		m_width  = width;
		m_height = height;
		m_multisample = multisample;

		if (m_multisample && (colorAttachmentType == atTexture || depthAttachmentType == atTexture))
			FATAL_ERROR("Framebuffer: texture attachments for multisample framebuffers not supported");

		// create and bind framebuffer
		glGenFramebuffers(1, &m_fbo);
		glBindFramebuffer(GL_FRAMEBUFFER, m_fbo);

		// create color attachment - either renderbuffer or texture (or none)
		if      (colorAttachmentType == atTexture) { for (int i = 0; i < numColorAttachments; i++) createColorTextureAttachment(i); }
		else if (colorAttachmentType == atBuffer)  createColorRenderbufferAttachment();

		// create color attachment - either renderbuffer or texture (or none)
		if      (depthAttachmentType == atTexture) createDepthTextureAttachment();
		else if (depthAttachmentType == atBuffer)  createDepthRenderbufferAttachment();

		// explicitly state if there is no color attachment
		if (colorAttachmentType == atNone) {
			glDrawBuffer(GL_NONE);
			glReadBuffer(GL_NONE);
		}

		if (numColorAttachments > 1) {
			GLuint bufs[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3, GL_COLOR_ATTACHMENT4, GL_COLOR_ATTACHMENT5, GL_COLOR_ATTACHMENT6, GL_COLOR_ATTACHMENT7 };
			glDrawBuffers(numColorAttachments, bufs);
		}

		// check if everything is ok, then unbind
		GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
		if (status != GL_FRAMEBUFFER_COMPLETE) {
			FATAL_ERROR("Framebuffer: creation failed, status code " << status << ". Called with w=" << width << " h=" << height << " cat=" << colorAttachmentType << "[" << colorTextureArraySize << "]  dat=" << depthAttachmentType << "[" << depthTextureArraySize << "] ms=" << multisample );
		}

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}

	/// Bind the framebuffer as active OpenGL framebuffer.
	inline void bind(bool setViewport = true) {
		glBindFramebuffer(GL_FRAMEBUFFER, m_fbo);
		if (setViewport) glViewport(0, 0, m_width, m_height);
	}

	// to be set before calling init(): (pixel format, pixel type are irrelevant)
	GLenum COLOR_INT_FORMAT[MAX_COLOR_ATTACHMENTS] = { GL_RGBA8,GL_RGBA8,GL_RGBA8,GL_RGBA8,GL_RGBA8,GL_RGBA8,GL_RGBA8,GL_RGBA8 };	///< Format for colour attachments
	GLenum DEPTH_INT_FORMAT = GL_DEPTH_COMPONENT24;	///< Format for depth attachment
	///@cond IGNORE
	GLenum COLOR_PIXEL_FORMAT = GL_RGBA,            COLOR_PIXEL_TYPE = GL_UNSIGNED_BYTE;
	GLenum DEPTH_PIXEL_FORMAT = GL_DEPTH_COMPONENT, DEPTH_PIXEL_TYPE = GL_FLOAT;
	///@endcond

	int colorTextureArraySize = 0; ///< Create colour textures as simple GL_TEXTURE_2D (0) or GL_TEXTURE_2D_ARRAY (> 0)
	int depthTextureArraySize = 0; ///< Create depth  texture  as simple GL_TEXTURE_2D (0) or GL_TEXTURE_2D_ARRAY (> 0)

	GLuint externalColorTexture[MAX_COLOR_ATTACHMENTS] = { 0 };	///< External colour texture handles. @see init()
	GLuint externalDepthRenderBuffer = 0;						///< External depth render buffer handle. @see init()
private:

	GLuint m_fbo = 0, m_colorRB = 0, m_depthRB = 0;
	int m_width, m_height, m_multisample;
	GLuint m_texColor[MAX_COLOR_ATTACHMENTS] = { 0 };
	GLuint m_texDepth = 0;

	inline void createColorTextureAttachment(int slot) {
		m_texColor[slot] = externalColorTexture[slot] ? externalColorTexture[slot] : makeTexture(true, slot); // create texture and attach it to fbo as color attachment #slot
		if (colorTextureArraySize == 0) glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + slot, GL_TEXTURE_2D, m_texColor[slot], 0);
		else                            glFramebufferTextureLayer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + slot, m_texColor[slot], 0, 0);
	}

	inline void createColorRenderbufferAttachment() {
		m_colorRB = makeRenderbuffer(true); // create a renderbuffer and attach it to fbo as color attachment #0
		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, m_colorRB);
	}

	inline void createDepthTextureAttachment() {		
		m_texDepth = makeTexture(false, 0); // create texture and attach it to fbo as depth attachment
		if (depthTextureArraySize == 0) glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, m_texDepth, 0);
		else                            glFramebufferTextureLayer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, m_texDepth, 0, 0);
	}

	inline void createDepthRenderbufferAttachment() {
		m_depthRB = externalDepthRenderBuffer ? externalDepthRenderBuffer : makeRenderbuffer(false); // create a renderbuffer and attach it to fbo as depth attachment
		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, m_depthRB);
	}

	inline GLuint makeTexture(bool color, int slot) {
		int arrayCnt;
		GLenum target;
		if (color && (colorTextureArraySize > 0)) {
			arrayCnt = colorTextureArraySize;
			target = GL_TEXTURE_2D_ARRAY;
		} else if ((!color) && (depthTextureArraySize > 0)) {
			arrayCnt = depthTextureArraySize;
			target = GL_TEXTURE_2D_ARRAY;
		} else {
			arrayCnt = 0;
			target = GL_TEXTURE_2D;
		}

		GLuint tex;
		glGenTextures(1, &tex);
		glBindTexture(target, tex);
		if      ( color && arrayCnt == 0) glTexImage2D(target, 0, COLOR_INT_FORMAT[slot], m_width, m_height,           0, COLOR_PIXEL_FORMAT, COLOR_PIXEL_TYPE, NULL);
		else if ( color && arrayCnt != 0) glTexImage3D(target, 0, COLOR_INT_FORMAT[slot], m_width, m_height, arrayCnt, 0, COLOR_PIXEL_FORMAT, COLOR_PIXEL_TYPE, NULL);
		else if (!color && arrayCnt == 0) glTexImage2D(target, 0, DEPTH_INT_FORMAT,       m_width, m_height,           0, DEPTH_PIXEL_FORMAT, DEPTH_PIXEL_TYPE, NULL);
		else                              glTexImage3D(target, 0, DEPTH_INT_FORMAT,       m_width, m_height, arrayCnt, 0, DEPTH_PIXEL_FORMAT, DEPTH_PIXEL_TYPE, NULL);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST); // no interpolation, no mipmapping - for now
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		return tex;
	}

	inline GLuint makeRenderbuffer(bool color) {
		GLuint rb;
		glGenRenderbuffers(1, &rb);
		glBindRenderbuffer(GL_RENDERBUFFER, rb);
		if (m_multisample) {
			if (color) glRenderbufferStorageMultisample(GL_RENDERBUFFER, m_multisample, COLOR_INT_FORMAT[0], m_width, m_height);
			else       glRenderbufferStorageMultisample(GL_RENDERBUFFER, m_multisample, DEPTH_INT_FORMAT,    m_width, m_height);
		} else {
			if (color) glRenderbufferStorage(GL_RENDERBUFFER, COLOR_INT_FORMAT[0], m_width, m_height);
			else       glRenderbufferStorage(GL_RENDERBUFFER, DEPTH_INT_FORMAT,    m_width, m_height);
		}
		return rb;
	}

};