// Portable file dialogs (include before Utils.h to avoid warnings)
#include "portable-file-dialogs.h"

#include "Utils.h"
#include "Shader.h"
#include "Mesh.h"
#include "Camera.h"

// filesystem for C++14 - Eigen has problems compiling with C++17...
#define _SILENCE_EXPERIMENTAL_FILESYSTEM_DEPRECATION_WARNING
#include <experimental/filesystem>

// Dear ImGui
#include "imgui/imgui.h"
#include "imgui/imgui_impl_glfw.h"
#include "imgui/imgui_impl_opengl3.h"

// ini file handling
#include "IniUtil.h"


// TODO!!
// ...
// still tex "shadows", even when clamped -> mipmap issue, not occaring without them
// custom gradient or gradient via texture
// MSAA
//
// Doxygen: private members

//#define USE_OPENGL_DEBUG_CALLBACK _DEBUG
#define USE_OPENGL_DEBUG_CALLBACK 1

//extern int meshmath_test(int argc, char** argv);

// OpenGL version we want to use (4.3 core)
const int opengl_version_major = 4;
const int opengl_version_minor = 3;
const char* glsl_version = "#version 430 core";

const int initial_win_width  = 1024; //1280;
const int initial_win_height =  768; // 960;

// globals
GLFWwindow* gWindow;
int gScreenWidth, gScreenHeight;
bool gWindowSizeChanged = false;
Mesh gMesh;
Camera gCamera;
bool gWireframe = false;
bool gCullFace = true;
bool gMultisampling = false;

bool usePerspective = false;
bool link_iso_line_params = true;

float mouse_scroll_y = 0.0f;

bool gui_active = true;
bool gui_wants_mouse = false;
bool gui_wants_keys  = false;
bool gui_dev_mode = false; // TODO: set to false for submission

bool file_dialog_active = false;

const char *gradientTexFn[] = { "grad_WhBu.jpg", "grad_WhGn.png", "grad_OrRd.png", "grad_magma.png" };
const char *gradientTexNames = "White-Blue\0White-Green\0Orange-Red\0Magma\0";
int gradientTex = 0;

// decl
void initWindow();
void setupOpenGL();
void key_callback(GLFWwindow* win, int key, int scancode, int action, int mods);
void handleInput();
void loadMeshAndInitCamera(std::string filename);
bool loadSettings(std::string fullFilename);
bool saveSettings(std::string fullFilename, bool overwrite);


// ImGui help marker (taken from ImGui demo)
static void HelpMarker(const char* desc)
{
	ImGui::TextDisabled("(?)");
	if (ImGui::IsItemHovered())
	{
		ImGui::BeginTooltip();
		ImGui::PushTextWrapPos(ImGui::GetFontSize() * 35.0f);
		ImGui::TextUnformatted(desc);
		ImGui::PopTextWrapPos();
		ImGui::EndTooltip();
	}
}

int main(int argc, char** argv) {
	//meshmath_test(argc, argv);

	initGlobals();
	setupOpenGL();

	// load shaders
	//new Shader("mesh",		{ "mesh.vert",		"mesh.frag" });
	new Shader("final",			{ "quad.vert",		"final.frag" });
	new Shader("sphere",		{ "sphere.vert",	"sphere.frag" });
	new Shader("contour",		{ "contour.vert",	"contour.geom",			"contour.frag" });
	new Shader("contour_line",	{ "contour.vert",	"contour_line.geom",	"contour.frag" });
	new Shader("jfa",			{ "quad.vert",		"jfa.frag" });
	new Shader("debug",			{ "quad.vert",		"debug.frag" });
	new Shader("render_tex",	{ "quad.vert",		"quad.frag" });


	// init camera
	gCamera.m_fov    = glm::radians(60.0f);
	gCamera.m_near   = 0.01f;
	gCamera.m_far    = 1000.0f;
	gCamera.m_aspect = (float)gScreenWidth / (float)gScreenHeight;
	gCamera.setRad(6.0f);
	gCamera.storeResetParameters();

	// init mesh
	gMesh.init(gScreenWidth, gScreenHeight);

	std::string filename = (argc > 1) ? argv[1] : gAssetsDir + "/example_meshes/vesselTree4.off";
	loadMeshAndInitCamera(filename);

	glViewport(0, 0, gScreenWidth, gScreenHeight);		// set viewport to cover the full window
	glClearColor(.2f, .2f, .2f, 0.0f);	    // set clear color to dark grey
	glClearColor((float)0x2C / 255.0f, (float)0x00 / 255.0f, (float)0x1E / 255.0f, 0.0f); // 2C001E
	glEnable(GL_DEPTH_TEST);	// enable depth test
	glEnable(GL_CULL_FACE);
	if (gMultisampling) glEnable(GL_MULTISAMPLE); else glDisable(GL_MULTISAMPLE);

	double t0 = glfwGetTime();
	while (!glfwWindowShouldClose(gWindow)) {
		double t = glfwGetTime();
		double frameTime = t - t0;
		t0 = t;

		// handle window resizing
		if (gWindowSizeChanged) {
			gWindowSizeChanged = false;
			int w, h;
			glfwGetFramebufferSize(gWindow, &w, &h);
			if (w > 0 && h > 0) {
				gScreenWidth = w; gScreenHeight = h;
				glViewport(0, 0, gScreenWidth, gScreenHeight);
				gMesh.viewportChanged(gScreenWidth, gScreenHeight);
				gCamera.m_aspect = (float)gScreenWidth / (float)gScreenHeight;
			}
		}

		if (gui_active) {
			ImGui_ImplOpenGL3_NewFrame();
			ImGui_ImplGlfw_NewFrame();
			ImGui::NewFrame();
			ImGuiIO& io = ImGui::GetIO();
			gui_wants_mouse = io.WantCaptureMouse;
			gui_wants_keys  = io.WantCaptureKeyboard || io.WantTextInput;
		} else {
			gui_wants_keys = gui_wants_mouse = false;
		}


		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glm::mat4 matV = gCamera.getViewMatrix();
		//glm::mat4 matP = usePerspective ? glm::perspective(gCamera.m_fov, gCamera.m_aspect, gCamera.m_near, gCamera.m_far)
		//                                : glm::ortho(-gScreenWidth * orthof, gScreenWidth * orthof, -gScreenHeight * orthof, gScreenHeight * orthof, gCamera.m_near, gCamera.m_far);

		glm::mat4 matP;
		if (usePerspective) {
			matP = glm::perspective(gCamera.m_fov, gCamera.m_aspect, gCamera.m_near, gCamera.m_far);
		} else {
			//float orthof = gCamera.getRad() * 0.001f;
			float orthof = gCamera.getRad() * 0.5f; // 200.f;//  0.1f;
			float cam_near = 0.01f, cam_far = 1000.f;
			float d = glm::length(gMesh.getCenter() - gCamera.getWorldPosition());
			cam_near = d - 2.f * gMesh.getRadius(); // * 2 just to be on the safe side
			cam_far  = d + 2.f * gMesh.getRadius(); // ditto
			//printf("near %f , far %f\n", cam_near, cam_far);

			matP = glm::ortho(-orthof * gCamera.m_aspect, orthof  * gCamera.m_aspect, -orthof, orthof, cam_near, cam_far);
		}


		glPolygonMode(GL_FRONT_AND_BACK, gWireframe ? GL_LINE : GL_FILL);
		if (gCullFace) glEnable(GL_CULL_FACE); else glDisable(GL_CULL_FACE);

		gMesh.time = fmod((float)glfwGetTime(), 10.f);
		gMesh.render(matV, matP, !usePerspective);

		// GUI
		if (gui_active) {
			using namespace ImGui;

			static bool showDemo = false;
			Begin("Treelike");

			if (!gui_dev_mode) Text("F1: toggle GUI"); else Text("F1: toggle GUI, F2: developer controls");
			//Text("Frametime: %.1f ms", 1000.f / GetIO().Framerate);
			//Checkbox("Show developer controls", &gui_dev_mode);
			//if (gui_dev_mode) {
			//	SameLine();
			//	if (Button("ImGui info")) showDemo = true;
			//}
			Separator();

			Text("Main visualisations");

			Checkbox("Lighting",    &gMesh.useBlinnPhong);
			Checkbox("Endpoints",   &gMesh.renderEndpoints);
			Checkbox("Root point",  &gMesh.renderRootpoint);
			Checkbox("Scalar field##EnableScalarField", &gMesh.scalar_field.enabled);
			if (gMesh.scalar_field.enabled) {
				SameLine(); Checkbox("on contour##ScalarFieldOnContour", &gMesh.scalar_field.onContour);
			}

			Checkbox("U##EnableIsoU",          &gMesh.isolines.enabled[0]); SameLine();
			Checkbox("V isolines##EnableIsoV", &gMesh.isolines.enabled[1]);
			Checkbox("Hatching##EnableHatching", &gMesh.hatching.enabled);
			Checkbox("Decals##EnableDecals", &gMesh.decals.enabled);
			if (gMesh.decals.enabled) {
				SameLine(); Checkbox("on contour##DecalsOnContour", &gMesh.decals.onContour);
			}

			Separator();
			Text("Parameters");

			SliderFloat("Cont.width", &gMesh.contourWidth, 1.0f, 10.0f, "%.1f");

			if (CollapsingHeader("Light,Material,Color")) {
				InputFloat3("Light dir", &gMesh.dirLight.direction.x);
				Checkbox("Viewer-relative light", &gMesh.dirLight.viewerRelative);
				InputFloat("Ambient", &gMesh.ambientLight);
				InputFloat("Diff",  &gMesh.material.diffuse);
				InputFloat("Spec",  &gMesh.material.specular);
				InputFloat("Shiny", &gMesh.material.shiny);

				InputFloat("Sphere size", &gMesh.sphereSize);

				ColorEdit3("Background##bgCol",       &gMesh.colorBackground.x, ImGuiColorEditFlags_NoInputs | ImGuiColorEditFlags_NoLabel); SameLine();
				ColorEdit3("Contour##contCol",        &gMesh.colorContour.x,    ImGuiColorEditFlags_NoInputs | ImGuiColorEditFlags_NoLabel); SameLine();
				ColorEdit3("Mesh##meshCol",           &gMesh.colorMesh.x,       ImGuiColorEditFlags_NoInputs | ImGuiColorEditFlags_NoLabel); SameLine();
				ColorEdit3("MeshPlain##meshPlainCol", &gMesh.colorMeshPlain.x,  ImGuiColorEditFlags_NoInputs | ImGuiColorEditFlags_NoLabel); SameLine();
				Text("Backg./Cont./Mesh/Unlit color");
				Checkbox("Perspective", &usePerspective);
			}

			//if (gui_dev_mode && CollapsingHeader("Mesh ops")) {
			//	if (Button("Calc Endpoints")) gMesh.calcEndpoints();
			//	if (Button("Calc U")) gMesh.calcParamU();
			//}

			if (gui_dev_mode && CollapsingHeader("Developer")) {
				//Checkbox("Debug rend.", &gMesh.renderDebug);
				Checkbox("Depth aware JFA", &gMesh.depthAwareJFA);

				Checkbox("Add contours", &gMesh.addContours);
				InputFloat2("PolyOff", gMesh.polyOffset);
				//Combo("uvMode", &gMesh.uvMode, "None\0U param\0V param\0Hatching\0Tex\0");
				InputFloat("Debug float", &gMesh.debug_float, 0.f, 0.f, "%.5f");
				if (Button("ImGui info")) showDemo = true;
			}
			if (gMesh.scalar_field.enabled && CollapsingHeader("Scalar field##ScalarFieldParams")) {
				PushID("ScalarFieldStuff");
				if (Combo("Colour map", &gradientTex, gradientTexNames)) {
					gMesh.loadGradientTexture(gradientTexFn[gradientTex]);
				}
				Checkbox("Invert", &gMesh.scalar_field.invert);
				PopID();
			}
			if (gMesh.hatching.enabled && CollapsingHeader("Hatching##HatchingParams")) {
				PushID("HatchingStuff");
				SliderFloat("Frequency", &gMesh.hatching.f, 0.f, 500.f);		//SameLine(); HelpMarker("Frequency of hatch lines");
				SliderFloat("Offset",    &gMesh.hatching.o, 0.f, 20.f);
				SliderFloat("Width",     &gMesh.hatching.w, 0.f, 1.f);
				SliderFloat("Length",    &gMesh.hatching.l, 0.f, 10.f, "%.1f");
				InputFloat ("Strength",  &gMesh.hatching.s);
				SliderInt  ("Number",    &gMesh.hatching.n, 1, 10);
				PopID();
			}
			if ((gMesh.isolines.enabled[0] || gMesh.isolines.enabled[1]) && CollapsingHeader("Isolines##IsolinesParams")) {
				PushID("IsolinesStuff");
				Checkbox("U##EnableIsoU", &gMesh.isolines.enabled[0]); SameLine();
				Checkbox("V##EnableIsoV", &gMesh.isolines.enabled[1]); SameLine();
				Checkbox("Same parameters", &link_iso_line_params);
				if (link_iso_line_params) {
					gMesh.isolines.freq [1] = gMesh.isolines.freq [0];
					gMesh.isolines.width[1] = gMesh.isolines.width[0];
					SliderFloat("Freq",  gMesh.isolines.freq,  0.0f, 500.0f, "%.0f");
					SliderFloat("Width", gMesh.isolines.width, 0.0f, 10.0f,  "%.1f");
				} else {
					SliderFloat2("Freq",  gMesh.isolines.freq,  0.0f, 500.0f, "%.0f");
					SliderFloat2("Width", gMesh.isolines.width, 0.0f, 10.0f,  "%.1f");
				}
				PopID();
			}
			if (gMesh.decals.enabled && CollapsingHeader("Decals##DecalsParams")) {
				PushID("DecalsStuff");
				InputFloat3("Scale",  &gMesh.decals.scale.x);
				InputFloat2("Offset", &gMesh.decals.offset.x);
				Checkbox("Clamp V",   &gMesh.decals.clampV);	SameLine();
				Checkbox("Invert V",  &gMesh.decals.invertV);	SameLine();
				Checkbox("Anim",      &gMesh.decals.animate);
				PopID();
			}

			Separator();
			if (Button("Load mesh...")) {
				static std::string lastMeshFn = "";

				// windows file dialog (or pfd) is picky about slashes and full paths... convert to canonical path
				std::string def = (lastMeshFn != "") ? lastMeshFn : std::experimental::filesystem::canonical(gAssetsDir + "/example_meshes/").string();
				auto f = pfd::open_file("Load mesh", def, { "Model files (.obj .off)", "*.obj *.off", "All files", "*" });
				auto res = f.result();
				if (res.size()) {
					std::cout << "Loading mesh: " << res[0] << std::endl;
					loadMeshAndInitCamera(res[0]);
					lastMeshFn = res[0];
				}
			}

			static std::string lastParamsFn = "";

			if (Button("Load params...")) {
				if (lastParamsFn == "") lastParamsFn = std::experimental::filesystem::canonical(gAssetsDir + "/params/").string();
				auto fns = pfd::open_file("Load parameters", lastParamsFn, { "Parameter files (.ini)", "*.ini", "All files", "*" }).result();
				if (fns.size()) {
					loadSettings(fns[0]);
				}
			}
			SameLine();
			if (Button("Save params...")) {
				if (lastParamsFn == "") lastParamsFn = std::experimental::filesystem::canonical(gAssetsDir + "/params/").string();
				auto fn = pfd::save_file("Save parameters as", lastParamsFn, { "Parameter files (.ini)", "*.ini", "All files", "*" }).result();
				if (fn != "") {
					saveSettings(fn, true);
					lastParamsFn = fn;
				}
			}

			End();

			if (showDemo) ShowDemoWindow(&showDemo);

			Render();
			ImGui_ImplOpenGL3_RenderDrawData(GetDrawData());
		}


		glfwSwapBuffers(gWindow);
		glfwPollEvents();
		handleInput();

	}

	// cleanup

	gMesh.cleanup();
	Shader::deleteAllShaders();
	glfwTerminate();

	return EXIT_SUCCESS;
}

void loadMeshAndInitCamera(std::string filename) {
	gMesh.loadFromFile(filename);
	gMesh.calcEndpoints();
	gMesh.calcParamU();

	// set a good camera position if we didn't scale the mesh to standard size
	gCamera.reset();
	if (true || !gMesh.scaleMesh) {
		float r = gMesh.getRadius() / tan(gCamera.m_fov * 0.5f);
		r *= 0.75f; // a bit closer
		gCamera.setRad(r);
	}
	if (!gMesh.centerMesh) {
		gCamera.setFocus(gMesh.getCenter());
	}
}

// callbacks
void key_callback(GLFWwindow* win, int key, int scancode, int action, int mods) {
	if (gui_wants_keys && (key < GLFW_KEY_F1 || key > GLFW_KEY_F12)) return;

	const char* onoff[2] = { "off", "on" };

	if (action == GLFW_PRESS) {
		switch (key)
		{
		case GLFW_KEY_ESCAPE:
			glfwSetWindowShouldClose(gWindow, true);
			break;
		case GLFW_KEY_G:
		case GLFW_KEY_F1:
			gui_active = !gui_active;
			break;
		case GLFW_KEY_F2:
			gui_dev_mode = !gui_dev_mode;
			break;
		//case GLFW_KEY_F2:
		//	gCullFace = !gCullFace;
		//	PRINT_OUT("Face culling " << onoff[gCullFace]);
		//	break;
		//case GLFW_KEY_F3:
		//	gWireframe = !gWireframe;
		//	PRINT_OUT("Wireframe " << onoff[gWireframe]);
		//	break;
		case GLFW_KEY_M:
			gMultisampling = !gMultisampling;
			if (gMultisampling) glEnable(GL_MULTISAMPLE); else glDisable(GL_MULTISAMPLE);
			PRINT_OUT("Multisampling " << onoff[gMultisampling]);
			break;
		}
	}
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset) {
	if (gui_wants_mouse) return;

	mouse_scroll_y += (float)yoffset; // handled in handleInput();
}

void handleInput() {
	static float old_x = 0.0f, old_y = 0.0f;

	if (gui_wants_mouse) return;

	double dblx, dbly;
	float x, y, dx, dy;
	glfwGetCursorPos(gWindow, &dblx, &dbly); x = (float)dblx; y = (float)dbly;
	y = gScreenHeight - 1.0f - y; // make y point upwards
	dx = x - old_x;
	dy = y - old_y;
	old_x = x;
	old_y = y;

	constexpr float rotationSpeed = glm::radians(0.25f); // rotation per pixel
	const float zoomSpeed = 0.5f;
	const float strafeSpeed = 0.01f; // units per pixel
	float speedFactor = gMesh.getRadius() * 0.2f;


	bool shift = glfwGetKey(gWindow, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS || glfwGetKey(gWindow, GLFW_KEY_RIGHT_SHIFT) == GLFW_PRESS;

	if (glfwGetMouseButton(gWindow, GLFW_MOUSE_BUTTON_LEFT) == GLFW_PRESS) {
		gCamera.moveTheta(dy * rotationSpeed);
		gCamera.movePhi(dx * rotationSpeed);
	}
	if (glfwGetMouseButton(gWindow, GLFW_MOUSE_BUTTON_MIDDLE) == GLFW_PRESS || glfwGetMouseButton(gWindow, GLFW_MOUSE_BUTTON_RIGHT) == GLFW_PRESS) {
		gCamera.moveRightUp(-dx * strafeSpeed * speedFactor, -dy * strafeSpeed * speedFactor);
	}

	if (mouse_scroll_y != 0.0) {
		float spd = shift ? zoomSpeed * 5.0f : zoomSpeed;
		gCamera.moveRad(-mouse_scroll_y * spd * speedFactor);
		//PRINT_OUT("rad " << gCamera.getRad());
		mouse_scroll_y = 0.0;
	}

}

void framebuffer_size_callback(GLFWwindow* win, int width, int height) {
	// minimizing the window sets width and height to zero. Do NOT set windowSizeChanged in that case, or bad things will happen
	if (width <= 0 || height <= 0) return;
	gWindowSizeChanged = true;
}

/// Initialize window for rendering.
void initWindow() {
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, opengl_version_major);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, opengl_version_minor);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE); // core only
	glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // resizable
#if USE_OPENGL_DEBUG_CALLBACK
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE); // create a debug context (only for debug version)
#endif
	glfwWindowHint(GLFW_SAMPLES, 4); // <antialiasing>xMSAA antialiasing (default 4xMSAA)

	gWindow = glfwCreateWindow(initial_win_width, initial_win_height, "Treelike", nullptr, nullptr);
	glfwGetFramebufferSize(gWindow, &gScreenWidth, &gScreenHeight);

	glfwMakeContextCurrent(gWindow); // activate the context
	//glfwSwapInterval(0); // disable vsync

	// set callbacks
	glfwSetKeyCallback(gWindow, key_callback);
	glfwSetScrollCallback(gWindow, scroll_callback);
	glfwSetFramebufferSizeCallback(gWindow, framebuffer_size_callback);
}

void setupOpenGL() {
	// init GLFW
	if (!glfwInit()) FATAL_ERROR("Failed to init GLFW");

	// create window and init GL context
	initWindow();

	// init GLEW
	glewExperimental = true;
	if (GLEW_OK != glewInit()) FATAL_ERROR("Failed to init GLEW");

	// init debug callback
#if USE_OPENGL_DEBUG_CALLBACK
	glDebugMessageCallback(openGL_debugCallback, nullptr);
	// disable debug messages intended for RenderDoc
	glDebugMessageControl(GL_DEBUG_SOURCE_APPLICATION, GL_DEBUG_TYPE_PUSH_GROUP, GL_DEBUG_SEVERITY_NOTIFICATION, 0, NULL, GL_FALSE);
	glDebugMessageControl(GL_DEBUG_SOURCE_APPLICATION, GL_DEBUG_TYPE_POP_GROUP, GL_DEBUG_SEVERITY_NOTIFICATION, 0, NULL, GL_FALSE);
	glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS); // want synchronous callbacks
#endif

	// init imgui
	IMGUI_CHECKVERSION();
	ImGui::CreateContext();
	//ImGuiIO& io = ImGui::GetIO();
	//io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;
	ImGui::StyleColorsDark();
	ImGui_ImplGlfw_InitForOpenGL(gWindow, true);
	ImGui_ImplOpenGL3_Init(glsl_version);
}

bool saveSettings(std::string fullFilename, bool overwrite) {
	if (!overwrite) {
		struct stat buffer;
		if (stat(fullFilename.c_str(), &buffer) == 0) return false;
	}

	mINI::INIFile file(fullFilename);
	mINI::INIStructure ini;
	std::string sec;

	sec = "Settings";
	iniWriteBool (ini, sec, "useBlinnPhong",			gMesh.useBlinnPhong);
	iniWriteBool (ini, sec, "renderEndpoints",			gMesh.renderEndpoints);
	iniWriteBool (ini, sec, "renderRootpoint",			gMesh.renderRootpoint);
	iniWriteBool (ini, sec, "scalar_field.enabled",		gMesh.scalar_field.enabled);
	iniWriteBool (ini, sec, "scalar_field.onContour",	gMesh.scalar_field.onContour);
	iniWriteBool (ini, sec, "isolines.enabled.0",		gMesh.isolines.enabled[0]);
	iniWriteBool (ini, sec, "isolines.enabled.1",		gMesh.isolines.enabled[1]);
	iniWriteBool (ini, sec, "hatching.enabled",			gMesh.hatching.enabled);
	iniWriteBool (ini, sec, "decals.enabled",			gMesh.decals.enabled);
	iniWriteBool (ini, sec, "decals.onContour",			gMesh.decals.onContour);
	iniWriteFloat(ini, sec, "contourWidth",				gMesh.contourWidth);
	iniWriteVec3 (ini, sec, "dirLight.direction",		gMesh.dirLight.direction);
	iniWriteBool (ini, sec, "dirLight.viewerRelative",	gMesh.dirLight.viewerRelative);
	iniWriteFloat(ini, sec, "ambientLight",				gMesh.ambientLight);
	iniWriteFloat(ini, sec, "material.diffuse",			gMesh.material.diffuse);
	iniWriteFloat(ini, sec, "material.specular",		gMesh.material.specular);
	iniWriteFloat(ini, sec, "material.shiny",			gMesh.material.shiny);
	iniWriteFloat(ini, sec, "sphereSize",				gMesh.sphereSize);
	iniWriteVec3 (ini, sec, "colorBackground",			gMesh.colorBackground);
	iniWriteVec3 (ini, sec, "colorContour",				gMesh.colorContour);
	iniWriteVec3 (ini, sec, "colorMesh",				gMesh.colorMesh);
	iniWriteVec3 (ini, sec, "colorMeshPlain",			gMesh.colorMeshPlain);
	iniWriteBool (ini, sec, "usePerspective",			usePerspective);
	iniWriteBool (ini, sec, "depthAwareJFA",			gMesh.depthAwareJFA);
	iniWriteBool (ini, sec, "addContours",				gMesh.addContours);
	iniWriteFloat(ini, sec, "polyOffset_0",				gMesh.polyOffset[0]);
	iniWriteFloat(ini, sec, "polyOffset_1",				gMesh.polyOffset[1]);
	iniWriteFloat(ini, sec, "debug_float",				gMesh.debug_float);
	iniWriteInt  (ini, sec, "gradientTex",				gradientTex);
	iniWriteBool (ini, sec, "scalar_field.invert",		gMesh.scalar_field.invert);
	iniWriteFloat(ini, sec, "hatching.f",				gMesh.hatching.f);
	iniWriteFloat(ini, sec, "hatching.o",				gMesh.hatching.o);
	iniWriteFloat(ini, sec, "hatching.w",				gMesh.hatching.w);
	iniWriteFloat(ini, sec, "hatching.l",				gMesh.hatching.l);
	iniWriteFloat(ini, sec, "hatching.s",				gMesh.hatching.s);
	iniWriteInt  (ini, sec, "hatching.n",				gMesh.hatching.n);
	iniWriteBool (ini, sec, "link_iso_line_params",		link_iso_line_params);
	iniWriteFloat(ini, sec, "isolines.freq_0",			gMesh.isolines.freq [0]);
	iniWriteFloat(ini, sec, "isolines.freq_1",			gMesh.isolines.freq [1]);
	iniWriteFloat(ini, sec, "isolines.width_0",			gMesh.isolines.width[0]);
	iniWriteFloat(ini, sec, "isolines.width_1",			gMesh.isolines.width[1]);
	iniWriteVec3 (ini, sec, "decals.scale",				gMesh.decals.scale);
	iniWriteVec2 (ini, sec, "decals.offset",			gMesh.decals.offset);
	iniWriteBool (ini, sec, "decals.clampV",			gMesh.decals.clampV);
	iniWriteBool (ini, sec, "decals.invertV",			gMesh.decals.invertV);
	iniWriteBool (ini, sec, "decals.animate",			gMesh.decals.animate);

	// TODO: Save camera too?
	// Do NOT save mesh filename

	return file.generate(ini);
}

bool loadSettings(std::string fullFilename) {
	mINI::INIFile file(fullFilename);
	mINI::INIStructure ini;

	if (!file.read(ini)) return false;

	std::string sec;

	sec = "Settings";
	iniReadBool  (ini, sec, "useBlinnPhong",			gMesh.useBlinnPhong);
	iniReadBool  (ini, sec, "renderEndpoints",			gMesh.renderEndpoints);
	iniReadBool  (ini, sec, "renderRootpoint",			gMesh.renderRootpoint);
	iniReadBool  (ini, sec, "scalar_field.enabled",		gMesh.scalar_field.enabled);
	iniReadBool  (ini, sec, "scalar_field.onContour",	gMesh.scalar_field.onContour);
	iniReadBool  (ini, sec, "isolines.enabled.0",		gMesh.isolines.enabled[0]);
	iniReadBool  (ini, sec, "isolines.enabled.1",		gMesh.isolines.enabled[1]);
	iniReadBool  (ini, sec, "hatching.enabled",			gMesh.hatching.enabled);
	iniReadBool  (ini, sec, "decals.enabled",			gMesh.decals.enabled);
	iniReadBool  (ini, sec, "decals.onContour",			gMesh.decals.onContour);
	iniReadFloat (ini, sec, "contourWidth",				gMesh.contourWidth);
	iniReadVec3  (ini, sec, "dirLight.direction",		gMesh.dirLight.direction);
	iniReadBool  (ini, sec, "dirLight.viewerRelative",	gMesh.dirLight.viewerRelative);
	iniReadFloat (ini, sec, "ambientLight",				gMesh.ambientLight);
	iniReadFloat (ini, sec, "material.diffuse",			gMesh.material.diffuse);
	iniReadFloat (ini, sec, "material.specular",		gMesh.material.specular);
	iniReadFloat (ini, sec, "material.shiny",			gMesh.material.shiny);
	iniReadFloat (ini, sec, "sphereSize",				gMesh.sphereSize);
	iniReadVec3  (ini, sec, "colorBackground",			gMesh.colorBackground);
	iniReadVec3  (ini, sec, "colorContour",				gMesh.colorContour);
	iniReadVec3  (ini, sec, "colorMesh",				gMesh.colorMesh);
	iniReadVec3  (ini, sec, "colorMeshPlain",			gMesh.colorMeshPlain);
	iniReadBool  (ini, sec, "usePerspective",			usePerspective);
	iniReadBool  (ini, sec, "depthAwareJFA",			gMesh.depthAwareJFA);
	iniReadBool  (ini, sec, "addContours",				gMesh.addContours);
	iniReadFloat (ini, sec, "polyOffset_0",				gMesh.polyOffset[0]);
	iniReadFloat (ini, sec, "polyOffset_1",				gMesh.polyOffset[1]);
	iniReadFloat (ini, sec, "debug_float",				gMesh.debug_float);
	iniReadInt   (ini, sec, "gradientTex",				gradientTex);
	iniReadBool  (ini, sec, "scalar_field.invert",		gMesh.scalar_field.invert);
	iniReadFloat (ini, sec, "hatching.f",				gMesh.hatching.f);
	iniReadFloat (ini, sec, "hatching.o",				gMesh.hatching.o);
	iniReadFloat (ini, sec, "hatching.w",				gMesh.hatching.w);
	iniReadFloat (ini, sec, "hatching.l",				gMesh.hatching.l);
	iniReadFloat (ini, sec, "hatching.s",				gMesh.hatching.s);
	iniReadInt   (ini, sec, "hatching.n",				gMesh.hatching.n);
	iniReadBool  (ini, sec, "link_iso_line_params",		link_iso_line_params);
	iniReadFloat (ini, sec, "isolines.freq_0",			gMesh.isolines.freq [0]);
	iniReadFloat (ini, sec, "isolines.freq_1",			gMesh.isolines.freq [1]);
	iniReadFloat (ini, sec, "isolines.width_0",			gMesh.isolines.width[0]);
	iniReadFloat (ini, sec, "isolines.width_1",			gMesh.isolines.width[1]);
	iniReadVec3  (ini, sec, "decals.scale",				gMesh.decals.scale);
	iniReadVec2  (ini, sec, "decals.offset",			gMesh.decals.offset);
	iniReadBool  (ini, sec, "decals.clampV",			gMesh.decals.clampV);
	iniReadBool  (ini, sec, "decals.invertV",			gMesh.decals.invertV);
	iniReadBool  (ini, sec, "decals.animate",			gMesh.decals.animate);


	// re-load the gradient texture, may be a different one now
	gMesh.loadGradientTexture(gradientTexFn[gradientTex]);


	return true;
}