#include "Mesh.h"
#include "meshmath.h"
#include "stb_image_impl.h"

#include <igl/read_triangle_mesh.h>
#include <igl/per_vertex_normals.h>

glm::vec3 EigenRowVec3dToGlmVec3(const Eigen::RowVector3d &e) {
	return glm::vec3((float)e(0), (float)e(1), (float)e(2));
}

Mesh::Mesh() {
	dirLight.direction = glm::normalize(glm::vec3(0, -1, -1));
	dirLight.viewerRelative = false;
	ambientLight       = 0.2f;
	material.diffuse   = 0.7f;
	material.specular  = 0.3f;
	material.shiny     = 9.0f;

}

Mesh::~Mesh() {
	cleanup();
}

void Mesh::init(int viewportWidth, int viewportHeight) {
	if (inited) return;

	this->viewportWidth  = viewportWidth;
	this->viewportHeight = viewportHeight;

	final_shader        = Shader::findShader("final");
	sphere_shader       = Shader::findShader("sphere");
	contour_shader      = Shader::findShader("contour");
	contour_line_shader = Shader::findShader("contour_line");
	jfa_shader          = Shader::findShader("jfa");
	debug_shader        = Shader::findShader("debug");
	renderTex_shader	= Shader::findShader("render_tex");

	// --- init sphere
	Eigen::Matrix<float, Eigen::Dynamic, 3, Eigen::RowMajor> sphereV;
	Eigen::Matrix<int,   Eigen::Dynamic, 3, Eigen::RowMajor> sphereF;
	if (!igl::read_triangle_mesh(gAssetsDir + "/models/sphere.off", sphereV, sphereF))
		FATAL_ERROR("Failed to load sphere mesh");

	sphere.numIndices = (int)sphereF.size();

	glGenVertexArrays(1, &sphere.vao);
	glBindVertexArray(sphere.vao);

	// vertex positions
	glGenBuffers(1, &sphere.vbo);
	glBindBuffer(GL_ARRAY_BUFFER, sphere.vbo);
	glBufferData(GL_ARRAY_BUFFER, sphereV.size() * sizeof(float), sphereV.data(), GL_STATIC_DRAW);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), nullptr);
	glEnableVertexAttribArray(0);

	// index buffer
	glGenBuffers(1, &sphere.ebo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphere.ebo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sphereF.size() * sizeof(int), sphereF.data(), GL_STATIC_DRAW);

	// unbind
	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	// ----

	// generate an empty vao for fullscreen-quad rendering
	glGenVertexArrays(1, &dummy_vao);

	// init framebuffers
	recreateFramebuffers();

	// load textures
	loadTexture();
	loadGradientTexture("grad_WhBu.jpg");

	inited = true;
}

void Mesh::viewportChanged(int viewportWidth, int viewportHeight) {
	this->viewportWidth = viewportWidth;
	this->viewportHeight = viewportHeight;
	recreateFramebuffers();
}

void Mesh::cleanup() {
	clearCurrentMesh();
	if (inited) {
		glDeleteTextures(1, &texDecal);
		glDeleteTextures(1, &texGrad);

		glDeleteVertexArrays(1, &sphere.vao);
		glDeleteBuffers(1, &sphere.vbo);
		glDeleteBuffers(1, &sphere.ebo);

		glDeleteVertexArrays(1, &dummy_vao);

		cleanupFramebuffers();
	}
	inited = false;
}

void Mesh::clearCurrentMesh() {
	V = Eigen::MatrixXd(0, 0);
	F = Eigen::MatrixXi(0, 0);

	ixEndpoints.clear();
	ixRootPoint = 0;
	haveParam_U = false;

	glDeleteVertexArrays(1, &vao);
	glDeleteBuffers(3, vbo);
	glDeleteBuffers(1, &ebo);
}

void Mesh::initVao() {
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	glGenBuffers(3, vbo);

	// vertex positions
	Eigen::MatrixXf Vt = V.transpose().cast<float>();
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, Vt.size() * sizeof(float), Vt.data(), GL_STATIC_DRAW);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), nullptr);
	glEnableVertexAttribArray(0);

	// vertex normals
	Eigen::MatrixXd N;
	igl::per_vertex_normals(V, F, N);
	Eigen::MatrixXf Nt = N.transpose().cast<float>();
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, Nt.size() * sizeof(float), Nt.data(), GL_STATIC_DRAW);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), nullptr);
	glEnableVertexAttribArray(1);

	// uv coordinates
	glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
	//glBufferData(GL_ARRAY_BUFFER, Nt.size() * sizeof(float), Nt.data(), GL_STATIC_DRAW);
	copyUVtoVBO();
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(float), nullptr);
	glEnableVertexAttribArray(2);


	// index buffer
	Eigen::MatrixXi Ft = F.transpose();
	glGenBuffers(1, &ebo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, Ft.size() * sizeof(int), Ft.data(), GL_STATIC_DRAW);

	// unbind
	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0); // unbind ebo AFTER vao!
}

void Mesh::copyUVtoVBO() {
	// copy UV params to currently bound VBO
	Eigen::Matrix<float, Eigen::Dynamic, 2, Eigen::RowMajor> UV(V.rows(),2);
	UV.setZero();
	if (haveParam_U) UV.col(0) = paramU.cast<float>();
	if (haveParam_Q) UV.col(1) = paramQ.cast<float>();
	glBufferData(GL_ARRAY_BUFFER, UV.size() * sizeof(float), UV.data(), GL_STATIC_DRAW);
}

void Mesh::cleanupFramebuffers() {
	DELETE_AND_NULL(fbFinal);
	DELETE_AND_NULL(fbDebug);
	DELETE_AND_NULL(fbGBuffers);
	DELETE_AND_NULL(fbJFA[1]);
	DELETE_AND_NULL(fbJFA[0]);
}

void Mesh::recreateFramebuffers() {
	// create framebuffers for JFA and G-buffers
	cleanupFramebuffers();

	for (int i = 0; i < 2; i++) {
		fbJFA[i] = new Framebuffer();
		fbJFA[i]->COLOR_INT_FORMAT[0] = GL_RGBA32F; // store 32 bit floats per channel
		fbJFA[i]->init(viewportWidth, viewportHeight, Framebuffer::atTexture, Framebuffer::atNone); // JFA doesn't need a depth buffer
	}

	// G-Buffer has 3 attachments
	fbGBuffers = new Framebuffer();
	fbGBuffers->externalColorTexture[0] = fbJFA[0]->getColorTexture();	// #0 is same as fbJFA[0] attachment
	fbGBuffers->COLOR_INT_FORMAT[1] = GL_RGBA32F;						// #1 positions
	fbGBuffers->COLOR_INT_FORMAT[2] = GL_RGBA32F;						// #2 normals
	fbGBuffers->init(viewportWidth, viewportHeight, Framebuffer::atTexture, Framebuffer::atBuffer, 0, 3); // don't need depth as texture, but need a depth buffer

	// TODO: change interpolation? (default is nearest)

	fbFinal = new Framebuffer();
	fbFinal->externalDepthRenderBuffer = fbGBuffers->getDepthRenderBuffer(); // use the same depth buffer as for G-Buffers
	fbFinal->COLOR_INT_FORMAT[0] = GL_RGBA32F;
	fbFinal->init(viewportWidth, viewportHeight, Framebuffer::atTexture, Framebuffer::atBuffer);


	// for debugging...
	fbDebug = new Framebuffer();
	fbDebug->COLOR_INT_FORMAT[0] = GL_RGBA32F;
	fbDebug->init(viewportWidth, viewportHeight, Framebuffer::atTexture, Framebuffer::atBuffer);

}

void Mesh::loadTexture() {
	int w, h, comp;
	uint8_t* data = stbi_load((gAssetsDir + "/textures/arrow.png").c_str(), &w, &h, &comp, STBI_rgb_alpha);
	MYASSERT(comp == 4, "Invalid texture (no alpha)");

	glGenTextures(1, &texDecal);
	glBindTexture(GL_TEXTURE_2D, texDecal);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glGenerateMipmap(GL_TEXTURE_2D);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Mesh::loadGradientTexture(std::string filename) {
	if (texGrad) glDeleteTextures(1, &texGrad);

	int w, h, comp;
	uint8_t* data = stbi_load((gAssetsDir + "/textures/" + filename).c_str(), &w, &h, &comp, STBI_rgb);
	MYASSERT(comp == 3, "Invalid texture (not rgb)");

	glGenTextures(1, &texGrad);
	glBindTexture(GL_TEXTURE_2D, texGrad);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D, 0);

}

bool Mesh::loadFromFile(std::string filename) {
	clearCurrentMesh();

	if (!igl::read_triangle_mesh(filename, V, F)) {
		PRINT_ERROR("Failed to load mesh");
		return false;
	}

	cleanup_mesh(V, F); // remove duplicate vertices

	initVao();

	// center bounding box at origin and scale to approx. 5x5 cube
	Eigen::RowVectorXd minV = V.colwise().minCoeff();
	Eigen::RowVectorXd maxV = V.colwise().maxCoeff();
	Eigen::RowVectorXd midV = (minV + maxV) * 0.5;
	double maxSide = (maxV - minV).maxCoeff();
	meshScale = 5.0f / (float)maxSide;
	meshCenter = EigenRowVec3dToGlmVec3(midV);

	if (!centerMesh) midV.setZero();
	if (!scaleMesh)  meshScale = 1.0f;

	matM = glm::translate(glm::scale(glm::mat4(1), glm::vec3(meshScale) ), -glm::vec3(midV(0), midV(1), midV(2)));
	matNrmM = glm::mat3(glm::transpose(glm::inverse(matM)));

	Eigen::RowVectorXd tmp = (maxV - minV) * meshScale;
	meshRadius = float(sqrt(tmp.dot(tmp))) * 0.5f;
	printf("Mesh radius = %f, center is (%f , %f , %f)\n", meshRadius, meshCenter.x, meshCenter.y, meshCenter.z);

	return true;
}

void Mesh::calcEndpoints() {
	// TODO: invalidate all endpoint dependent stuff
	ixEndpoints.clear();
	ixRootPoint = 0;

	Eigen::MatrixXd Vskel(V);
	skeleton_extraction(Vskel, F, 5.0, 10);
	find_endpoints(Vskel, F, &ixEndpoints);
	ixRootPoint = find_root_point(ixEndpoints, V, F);
}

void Mesh::calcParamU() {
	calculate_U(V, F, ixRootPoint, paramU, nullptr);
	maxParamU = paramU.maxCoeff();
	haveParam_U = true;
	printf("maxU = %f\n", maxParamU);

	//// experiment: store min curvature in paramQ
	//MYASSERT(paramQ.rows() == paramU.rows(), "row mismatch");
	//paramQ = paramQ.cwiseAbs();
	//haveParam_Q = true;
	////dump_matrix("Q", paramQ);

	// update VBO[2]
	glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
	copyUVtoVBO();
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void Mesh::render(glm::mat4 &matV, glm::mat4 &matP, bool isOrthoProjection) {
	MYASSERT(inited, "Mesh not inited");

	Shader *sh;

	pushFramebuffer();

	// render mesh with contour to G-Buffers
	LABELED_SECTION_BEGIN("Draw G-Buffers");
	fbGBuffers->bind();
	// clear textures
	//GLuint bufPosNrm[] = { GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2 };	// positions, normals
	GLuint bufPos[]    = { GL_COLOR_ATTACHMENT1 };	// positions
	GLuint bufNrm[]    = { GL_COLOR_ATTACHMENT2 };	// normals
	GLuint bufJFAin[]  = { GL_COLOR_ATTACHMENT0 };	// JFA input
	GLuint bufAll[]    = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2 };	// all
	//glDrawBuffers(2, bufPosNrm);
	glDrawBuffers(1, bufPos);
	glClearColor(0, 0, 0, -1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffers(1, bufNrm);
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffers(1, bufJFAin);
	glClearColor(-1, -1, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffers(3, bufAll);
	glEnable(GL_DEPTH_TEST);	// enable depth test and depth writes
	glDepthMask(true);

	// render either base geometry only (addContours == false) or geometry + contour seam
	sh = contour_shader;
	sh->use();
	sh->setMat4 ("matMV", matV * matM);
	sh->setMat3 ("matNormalMV", glm::mat3(matV * glm::mat4(matNrmM)));
	sh->setMat4 ("matP", matP);
	sh->setBool ("orthoProjection", isOrthoProjection);
	sh->setFloat("contourSize", contourWidth * meshRadius * 0.002f);
	sh->setBool ("doContour", addContours);
	sh->setBool ("contour_lines_only", false); // TODO: remove, for debugging only

	if (!addContours) {
		glPolygonOffset(polyOffset[0], polyOffset[1]);	// draw polys a little bit offset, so there is no z-fighting with lines
		glEnable(GL_POLYGON_OFFSET_FILL);
	}

	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, (int)F.rows() * 3, GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);

	glDisable(GL_POLYGON_OFFSET_FILL);

	if (!addContours) {
		// if no visible contour is desired, draw contour lines for JFA only
		// only enable the output attachment for JFA
		glDrawBuffers(1, bufJFAin);
		glDepthFunc(GL_LEQUAL); // use less-or-equal depth comparison, so we overwrite triangle parts with same depth // TODO: doesn't work as expected...

		sh = contour_line_shader;
		sh->use();
		sh->setMat4 ("matMV", matV * matM);
		sh->setMat3 ("matNormalMV", glm::mat3(matV * glm::mat4(matNrmM)));
		sh->setMat4 ("matP", matP);
		sh->setBool ("orthoProjection", isOrthoProjection);
		sh->setBool ("contour_lines_only", true); // TODO: remove, for debugging only
		glBindVertexArray(vao);
		glDrawElements(GL_TRIANGLES, (int)F.rows() * 3, GL_UNSIGNED_INT, 0);
		glBindVertexArray(0);

		glDepthFunc(GL_LESS);
	}

	LABELED_SECTION_END;

	// TODO: polygon offset works - however messes up the depth trick in JFA a bit...


	// jump flooding
	LABELED_SECTION_BEGIN("Jump flood");
	int numSteps = (int)ceil(log2(std::max(viewportWidth, viewportHeight)));
	int jfaSrc = 0;
	glClearColor(0, 0, 0, 0);
	glDisable(GL_DEPTH_TEST);
	sh = jfa_shader;
	sh->use();
	for (int iStep = numSteps-1; iStep >= 0; iStep--) {

		// render to JFA[1], using texture from JFA[0] as input (in the first step)
		fbJFA[jfaSrc ^ 1]->bind();	// render to fbJFA[1]
		sh->setIVec2("screenSize", glm::ivec2(viewportWidth, viewportHeight));
		sh->setInt("stepSize", (int)pow(2, iStep));
		sh->setInt("texInput", 0);
		sh->setBool("depthAware", depthAwareJFA);
		sh->setBool("makeVtexture", iStep == 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, fbJFA[jfaSrc]->getColorTexture());

		//glClear(GL_COLOR_BUFFER_BIT);

		glBindVertexArray(dummy_vao);
		glDrawArraysInstanced(GL_TRIANGLES, 0, 3, 2);
		glBindVertexArray(0);

		jfaSrc ^= 1;
	}
	// jfaSrc now contains the V-texture
	GLuint vTextureHandle = fbJFA[jfaSrc]->getColorTexture();
	LABELED_SECTION_END;


	// draw final result

	//popFramebuffer();
	//if (renderDebug) {
	//	pushFramebuffer();
	//	fbDebug->bind();
	//}

	LABELED_SECTION_BEGIN("Final");
	fbFinal->bind();
	sh = final_shader;
	sh->use();
	sh->setVec3 ("dirLight_direction_VS", dirLight.viewerRelative ? glm::normalize(dirLight.direction) : glm::vec3(matV * glm::vec4(glm::normalize(dirLight.direction), 0)));
	sh->setFloat("ambientLight", ambientLight);
	sh->setFloat("mat_diffuse",  material.diffuse);
	sh->setFloat("mat_specular", material.specular);
	sh->setFloat("mat_shiny",    material.shiny);
	sh->setFloat("u_scale", abs(maxParamU) > 0 ? 1.0f / (float)maxParamU : 1.0f);
	//sh->setInt  ("uv_mode", uvMode);
	sh->setVec3 ("bgColor",      colorBackground);
	sh->setVec3 ("contourColor", colorContour);
	sh->setVec3 ("mesh_base_color", useBlinnPhong ? colorMesh : colorMeshPlain);
	sh->setBool ("hasContours",  addContours);
	sh->setIVec2("screenSize",   glm::ivec2(viewportWidth, viewportHeight));
	sh->setFloat("time",         time);
	sh->setFloat("debug_float",  debug_float);

	sh->setBool("use_blinn_phong", useBlinnPhong);

	sh->setBool ("scalar_field.enabled",    scalar_field.enabled);
	sh->setBool ("scalar_field.onContour",  scalar_field.onContour);
	sh->setBool ("scalar_field.useTexture", scalar_field.useTexture);
	sh->setBool ("scalar_field.invert",     scalar_field.invert);

	sh->setBool ("hatching.enabled", hatching.enabled);
	sh->setFloat("hatching.f",       hatching.f);
	sh->setFloat("hatching.o",       hatching.o);
	sh->setFloat("hatching.w",       hatching.w);
	sh->setFloat("hatching.l",       hatching.l);
	sh->setFloat("hatching.s",       hatching.s);
	sh->setInt  ("hatching.n",       hatching.n);

	for (int i = 0; i < 2; i++) {
		const std::string basename[2] = { "isolines_u", "isolines_v" };
		sh->setBool (basename[i] + ".enabled", isolines.enabled[i]);
		sh->setFloat(basename[i] + ".freq",    isolines.freq   [i]);
		sh->setFloat(basename[i] + ".width",   isolines.width  [i]);
	}

	sh->setBool ("decals.enabled",   decals.enabled);
	sh->setBool ("decals.onContour", decals.onContour);
	sh->setVec3 ("decals.scale",     decals.scale);
	sh->setVec2 ("decals.offset",    decals.offset);
	sh->setBool ("decals.clampV",    decals.clampV);
	sh->setBool ("decals.invertV",   decals.invertV);
	sh->setBool ("decals.animate",   decals.animate);

	sh->setInt  ("texPositionVS", 0);	glActiveTexture(GL_TEXTURE0);	glBindTexture(GL_TEXTURE_2D, fbGBuffers->getColorTexture(1));
	sh->setInt  ("texNormalVS",   1);	glActiveTexture(GL_TEXTURE1);	glBindTexture(GL_TEXTURE_2D, fbGBuffers->getColorTexture(2));
	sh->setInt  ("texParamV",     2);	glActiveTexture(GL_TEXTURE2);	glBindTexture(GL_TEXTURE_2D, vTextureHandle);
	sh->setInt  ("texDecal",      3);	glActiveTexture(GL_TEXTURE3);	glBindTexture(GL_TEXTURE_2D, texDecal);
	sh->setInt  ("texGrad",       4);	glActiveTexture(GL_TEXTURE4);	glBindTexture(GL_TEXTURE_2D, texGrad);

	glBindVertexArray(dummy_vao);
	glDrawArraysInstanced(GL_TRIANGLES, 0, 3, 2);
	glBindVertexArray(0);

	//// if debugging, render debug-fb to screen now
	//if (renderDebug) {
	//	popFramebuffer();

	//	sh = debug_shader;
	//	sh->use();
	//	sh->setInt  ("texFinal", 0);	glActiveTexture(GL_TEXTURE0);	glBindTexture(GL_TEXTURE_2D, fbDebug->getColorTexture());
	//	glBindVertexArray(dummy_vao);
	//	glDrawArraysInstanced(GL_TRIANGLES, 0, 3, 2);
	//	glBindVertexArray(0);
	//}

	// draw spheres
	if ((renderEndpoints || renderRootpoint) && ixEndpoints.size()) {
		LABELED_SECTION_BEGIN("Endpoint spheres");
		Shader* sh = sphere_shader;
		sh->use();
		glBindVertexArray(sphere.vao);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
		//glDisable(GL_DEPTH_TEST);
		//float scale = sphereSize / meshScale; // keep sphere size screen-relative
		float scale = meshRadius * sphereSize * 0.01;
		sh->setVec3 ("dirLight_direction_VS", dirLight.viewerRelative ? glm::normalize(dirLight.direction) : glm::vec3(matV * glm::vec4(glm::normalize(dirLight.direction), 0)));
		sh->setFloat("ambientLight", ambientLight);
		for (int i = 0; i < (int)ixEndpoints.size(); i++) {
			bool isRoot = (ixEndpoints[i] == ixRootPoint);
			if (!isRoot && !renderEndpoints) continue;

			glm::mat4 M = matM * glm::scale(glm::translate(glm::mat4(1), EigenRowVec3dToGlmVec3(V.row(ixEndpoints[i]))), glm::vec3(scale));
			sh->setMat4("matMVP", matP * matV * M);
			sh->setMat4("matMV",  matV * M);
			sh->setMat3("matNormalMV", glm::mat3(matV * glm::mat4(glm::mat3(glm::transpose(glm::inverse(M))))));

			sh->setVec4("color", isRoot && renderRootpoint ? glm::vec4(1, 0, 0, 0.5) : glm::vec4(0, 1, 0, 0.5));

			glDrawElements(GL_TRIANGLES, sphere.numIndices, GL_UNSIGNED_INT, 0);
		}
		glDisable(GL_BLEND);
		glBindVertexArray(0);
		LABELED_SECTION_END;
	}
	LABELED_SECTION_END;


	// render final framebuffer to screen
	popFramebuffer();

	LABELED_SECTION_BEGIN("Final to screen");
	sh = renderTex_shader;
	sh->use();
	sh->setInt  ("texFinal", 0);	glActiveTexture(GL_TEXTURE0);	glBindTexture(GL_TEXTURE_2D, fbFinal->getColorTexture());
	glBindVertexArray(dummy_vao);
	glDrawArraysInstanced(GL_TRIANGLES, 0, 3, 2);
	glBindVertexArray(0);
	LABELED_SECTION_END;
}

void Mesh::debugRender(glm::mat4& matV, glm::mat4& matP, const glm::vec3& eyePos) {
	Shader* sh = Shader::findShader("debug");
	sh->use();
	sh->setMat4 ("matM", matM);
	sh->setMat3 ("matNormalM", matNrmM);
	sh->setMat4 ("matMV", matV * matM);
	sh->setMat3 ("matNormalMV", glm::mat3(matV * glm::mat4(matNrmM)));
	sh->setMat4 ("matP", matP);
	sh->setVec3 ("eyePos", eyePos);
	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, (int)F.rows() * 3, GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);
}

