#pragma once

#include "Utils.h"
#include "Shader.h"
#include "Framebuffer.h"
#include "Eigen/Eigen"

/// Base class for a tree-like mesh.
///
/// This class encapsulates loading, preprocessing and rendering and is the main workhorse of the application.
class Mesh {
private:
	Eigen::MatrixXd V;	// Vertices
	Eigen::MatrixXi F;	// Faces
	std::vector<int> ixEndpoints;
	int ixRootPoint;

	Eigen::VectorXd paramU, paramQ;	// U- and a second parameterization (note: 2nd isn't the "real" V, but a placeholder for something else if needed)
	double maxParamU = 0.0;
	bool haveParam_U, haveParam_Q;

	// TODO: cache often used operators, like Laplacian

	bool inited = false;

	GLuint vbo[3] = { 0 };	// pos, normals, uv
	GLuint ebo = 0;			// index buffer
	GLuint vao = 0;

	struct {
		GLuint vbo = 0, ebo = 0, vao = 0;
		int numIndices = 0;	///< @private Why does Doxygen insist on trying to add this?
	} sphere; 

	GLuint dummy_vao = 0;

	Framebuffer* fbJFA[2] = { nullptr, nullptr };	// framebuffers for jump-flood-algorithm
	Framebuffer* fbGBuffers = nullptr;				// framebuffer for G-buffers
	Framebuffer* fbFinal = nullptr;					// framebuffer for final composition + endpoint spheres
	Framebuffer* fbDebug = nullptr;

	float meshScale = 1.0f;
	float meshRadius = 0.0f;
	glm::vec3 meshCenter = glm::vec3(0);

	glm::mat4 matM = glm::mat4(1);
	glm::mat3 matNrmM = glm::mat3(1);

	GLuint texDecal = 0;
	GLuint texGrad = 0;

	Shader *final_shader = nullptr;
	Shader *sphere_shader = nullptr;
	Shader *contour_shader = nullptr;
	Shader *contour_line_shader = nullptr;
	Shader *jfa_shader = nullptr;
	Shader *debug_shader = nullptr;
	Shader *renderTex_shader = nullptr;

	int viewportWidth  = 0;
	int viewportHeight = 0;


	void initVao();					///< Initialize OpenGL Vertex Array for rendering
	void clearCurrentMesh();		///< Clear the current mesh data
	void copyUVtoVBO();				///< Copy the calculated U-parameterisation to an OpenGL Vertex Buffer

	void cleanupFramebuffers();		///< Destroy internal framebuffers
	void recreateFramebuffers();	///< Create internal framebuffers

	void loadTexture();				///< Load decal texture

	void debugRender(glm::mat4 &matV, glm::mat4 &matP, const glm::vec3 &eyePos);
public:
	Mesh();
	~Mesh();

	/// Initialize mesh.
	///
	/// This must be called before using any other member functions.
	/// @param viewportWidth  Width of the viewport
	/// @param viewportHeight Height of the viewport
	/// @see cleanup()
	void init(int viewportWidth, int viewportHeight);

	/// Clean up and free allocated resources.
	///
	/// This should be called before destroying the Mesh instance.
	/// @see init()
	void cleanup();

	/// Load mesh data from a file.
	///
	/// Multiple file formats, including .obj and .off, are supported.
	/// @param   filename Filename, either absolute or relative to current working directory
	/// @returns True on success, false on failure
	bool loadFromFile(std::string filename);

	void calcEndpoints();	///< Calculate the endpoints of the mesh.
	void calcParamU();		///< Calculate the U-parameterisation of the mesh.

	/// Calculate V-parameterisation and render the mesh to the screen.
	///
	/// Rendering commences in multiple steps:
	/// - The mesh geometry is extended by contours and rendered to G-Buffers
	/// - Jump-flooding is performed to obtain the V-parameterisation
	/// - The final scene is rendered, using the G-Buffers and V-parameterisation
	/// - Endpoint spheres are overlaid
	///
	/// @param matV              Current view matrix
	/// @param matP              Current projection matrix
	/// @param isOrthoProjection True if matP is orthogonal, false for perspective projection
	void render(glm::mat4 &matV, glm::mat4 &matP, bool isOrthoProjection);

	/// Must be called when the viewport (window size) has been changed.
	/// @param viewportWidth  Width of the viewport
	/// @param viewportHeight Height of the viewport
	void viewportChanged(int viewportWidth, int viewportHeight);

	void loadGradientTexture(std::string filename); // TODO!! document
	
	float getRadius() { return meshRadius; }		///< Get radius of mesh bounding sphere.  @returns Radius
	glm::vec3 getCenter() { return meshCenter; }	///< Get centre of mesh bounding sphere.  @returns Centre coordinates


	/// A directional light source.
	struct DirLight {
		glm::vec3 direction;	///< Direction, originating *from* the light
		bool viewerRelative;	///< When true, direction is relative to the viewer (like a headlight).
	} dirLight;	///< Directional light parameters

	float ambientLight;			///< Ambient lighting.

	/// Material properties for Blinn-Phong shading.
	struct Material {
		float diffuse, specular, shiny;
	} material; ///< Mesh material parameters

	float sphereSize = 2.0f;	///< Sphere radius of endpoints.
								// (multiplied by mesh radius)

	bool centerMesh = false;	///< @private don't touch! - no longer used
	bool scaleMesh = false;		///< @private don't touch! - no longer used

	bool renderEndpoints = false;	///< Render all endpoints (including the root point)
	bool renderRootpoint = false;	///< Render the root point in red
	bool renderDebug = false;		///< @private debugging only
	//int uvMode = 0;


	bool depthAwareJFA = true;				///< Use depth-modification for JFA algorithm.
	bool addContours = true;				///< @private don't touch!
	float contourWidth = 3.0f;				///< Width of the contour.
	float polyOffset[2] = { 5.0f, 0.0f };	///< @private don't touch!

	glm::vec3 colorBackground = glm::vec3(0.5f);	///< Background colour of the viewer
	glm::vec3 colorContour    = glm::vec3(0);		///< Contour colour of the mesh
	glm::vec3 colorMesh       = glm::vec3(1,1,0);	///< Base colour of the mesh
	glm::vec3 colorMeshPlain  = glm::vec3(1,1,1);	///< Base colour of the mesh when no lighting is used

	bool useBlinnPhong = true; ///< Use blinn/phong lighting

	/// Parameters for drawing hatching lines.
	struct Hatching {
		bool enabled = false;	///< Toggle to enable/disable hatching
		float f = 200.0f;		///< Frequency
		float o = 0.0f;			///< Offset
		float w = 0.4f;			///< Width
		float l = 5.0f;			///< Length
		float s = 2.0f;		///< Strength
		int   n = 1;			///< Number of hatch strokes per hatch position
	} hatching; ///< Hatching parameters

	/// Parameters for drawing iso-level lines.
	///
	/// All parameters have two elements: [0] is for the U-direction, [1] is for the V-direction.
	struct IsoLines {
		bool enabled[2] = { false, false }; ///< Toggle to enable/disable iso lines
		float freq[2] = { 100.0f, 100.0f }; ///< Frequency
		float width[2] = { 1.0f, 1.0f };	///< Width
	} isolines; ///< Iso-line parameters

	/// Parameters for drawing decals, like arrows, on the mesh.
	struct Decals {
		bool enabled = false;						///< Toggle to enable/disable drawing decals
		bool onContour = false;						///< Draw on contour instead of on mesh
		glm::vec3 scale = glm::vec3(2.0f, 0.1f, 0.5f);	///< Scaling (.z is used instead of .y for contour-decals)
		glm::vec2 offset = glm::vec2(-0.5f, -0.1f); ///< Offset (applied after scaling)
		bool clampV  = true;						///< Clamp texture V to [0,1] after applying scaling and offset
		bool invertV = true;						///< Flip texture vertically (applied after clamping)
		bool animate = true;						///< Animate texture (Multiplies offset.x by time)
	} decals; ///< Decal parameters

	/// Parameters for scalar field visualisation.
	struct ScalarField {
		bool enabled = false;				///< Toggle to enable/disable scalar field
		bool onContour = false;				///< Colourise contour instead of mesh
		bool useTexture = true;				///< Use a texture for the scalar field colour range
		bool invert = true;					///< Invert colour range
		// TODO: colors
	} scalar_field; ///< Scalar-field parameters

	float time = 0.0f; ///< Current time (used for animating decals)

	float debug_float = 0.0f; ///< @private Debugging
};
