#include <sstream>
#include <fstream>
#include <vector>
#include <string>
#include "Utils.h"
#include "Shader.h"


// class Shader

std::vector<Shader *> Shader::m_AllShaders;
std::vector<std::string> Shader::m_customDefines;

#ifdef SHADER_PREVENT_BINDING_BOUND_SHADER
Shader *Shader::m_CurrentlyBoundShader = nullptr;
#endif

Shader::Shader(std::string name, std::vector<std::string> filenames, bool usesMVP) {
	m_name   = name;
	m_usesMVP = usesMVP;
	std::vector<std::string> varyings;
	m_handle = createShaderProgram(filenames, varyings);
	m_AllShaders.push_back(this);
}

Shader::Shader(std::string name, std::vector<std::string> filenames, std::vector<std::string> tfbVaryings, bool usesMVP) {
	m_name   = name;
	m_usesMVP = usesMVP;
	m_handle = createShaderProgram(filenames, tfbVaryings);
	m_AllShaders.push_back(this);
}

Shader::~Shader() {
#ifdef SHADER_PREVENT_BINDING_BOUND_SHADER
	if (this == Shader::m_CurrentlyBoundShader) Shader::m_CurrentlyBoundShader = nullptr;
#endif
	glDeleteProgram(m_handle);
	for (unsigned int i = 0; i < m_AllShaders.size(); i++) {
		if (m_AllShaders[i] == this) {
			m_AllShaders.erase(m_AllShaders.begin() + i);
			break;
		}
	}
}

Shader * Shader::findShader(std::string name) {
	for (unsigned int i = 0; i < m_AllShaders.size(); i++) {
		if (m_AllShaders[i]->getName().compare(name) == 0 ) {
			return m_AllShaders[i];
		}
	}
	FATAL_ERROR("Shader " << name << " is not loaded")
}

void Shader::deleteAllShaders() {
	while (!m_AllShaders.empty())
		delete m_AllShaders.back();
}


std::string Shader::loadShaderFile(std::string fileName)
{
	std::string	source;

	try {
		std::ifstream ifs;
		ifs.exceptions(std::ifstream::failbit | std::ifstream::badbit); // enable exceptions
		ifs.open(gAssetsDir + "/shaders/" + fileName);
		std::stringstream sstr;
		sstr << ifs.rdbuf();
		ifs.close();
		source = sstr.str();
	}
	catch (std::ifstream::failure e) {
		FATAL_ERROR("ERROR loading shader file \"" << fileName << "\"");
	}

	return source;
}

// Load a shader from a text file and compile, return handle
// Terminate on errors
GLuint Shader::loadAndCompileShader(std::string fileName, GLenum shaderType)
{
	std::string	source = loadShaderFile(fileName);

	// process #include ... very rough
	// must exactly look like (1 space after #include): #input "somefilename"
	// there is NO check for cyclic includes!
	while (true) {
		int idx = (int)source.find("#include");		// find next #include statement
		if (idx == std::string::npos) break;

		// extract include filename
		int idx1 = idx + 9;
		if (source[idx1] != '\"')
			FATAL_ERROR("ERROR: invalid #include syntax in shader (leading quotes missing) \"" << fileName << "\"");
		int idx2 = (int)source.find("\"", idx1 + 1);
		if (idx2 == std::string::npos)
			FATAL_ERROR("ERROR: invalid #include syntax in shader (ending quotes missing) \"" << fileName << "\"");
		std::string incfileName = source.substr(idx1 + 1, idx2 - idx1 - 1);

		// load include file
		std::string incsource = loadShaderFile(incfileName);

		// remove #include line and insert loaded file instead
		source.erase(idx, idx2 - idx + 1);
		source.insert(idx, incsource);
	}

	// process keywords
	{
		int idx = (int)source.find(SHADER_KEYWORD_INSERTDEFINES);
		if (idx != std::string::npos) {
			for (int i = (int)m_customDefines.size() - 1; i >= 0; i--)
				source.insert(idx, m_customDefines[i]);
		}
	}

	if (source.find(SHADER_KEYWORD_NEEDS_SHADOWMAP) != std::string::npos) m_needsShadowmap = true;

	GLuint handle = glCreateShader(shaderType);
	const char* codePtr = source.c_str();
	glShaderSource(handle, 1, &codePtr, nullptr);
	glCompileShader(handle);

	// check for compile errors
	GLint succeeded;
	glGetShaderiv(handle, GL_COMPILE_STATUS, &succeeded);
	if (succeeded == GL_FALSE) {
		GLint logSize;
		glGetShaderiv(handle, GL_INFO_LOG_LENGTH, &logSize);
		GLchar* log = new char[logSize];
		glGetShaderInfoLog(handle, logSize, nullptr, log);

		FATAL_ERROR("ERROR compiling shader \"" << fileName << "\": " << log);

		//delete[] log;	// never reached
	}

	return handle;
}

// Load shaders from text files, compile, link and return program handle
// Terminate on errors
GLuint Shader::createShaderProgram(std::vector<std::string> filenames, std::vector<std::string> tfbVaryings) {
	GLuint handle = glCreateProgram();

	std::vector<GLuint> shaderHandles;
	for (unsigned int i = 0; i < filenames.size(); i++) {
		GLenum type;
		if      (filenames[i].find(".vert") != std::string::npos || filenames[i].find(".vs")  != std::string::npos) type = GL_VERTEX_SHADER;
		else if (filenames[i].find(".frag") != std::string::npos || filenames[i].find(".fs")  != std::string::npos) type = GL_FRAGMENT_SHADER;
		else if (filenames[i].find(".tesc") != std::string::npos || filenames[i].find(".tcs") != std::string::npos) type = GL_TESS_CONTROL_SHADER;
		else if (filenames[i].find(".tese") != std::string::npos || filenames[i].find(".tes") != std::string::npos) type = GL_TESS_EVALUATION_SHADER;
		else if (filenames[i].find(".geom") != std::string::npos || filenames[i].find(".gs")  != std::string::npos) type = GL_GEOMETRY_SHADER;
		else if (filenames[i].find(".comp") != std::string::npos || filenames[i].find(".cs")  != std::string::npos) type = GL_COMPUTE_SHADER;
		else FATAL_ERROR("Unknown shader type: " << filenames[i]);
		shaderHandles.push_back(loadAndCompileShader(filenames[i], type));
		glAttachShader(handle, shaderHandles.back());
	}

	// set transform feedback varyings before linking
	if (tfbVaryings.size() > 0) {
		GLchar **vars = new GLchar * [tfbVaryings.size()];
		for (unsigned int i = 0; i < tfbVaryings.size(); i++) vars[i] = (GLchar *) tfbVaryings[i].c_str();
		glTransformFeedbackVaryings(handle, (GLsizei)tfbVaryings.size(), vars, GL_INTERLEAVED_ATTRIBS);
		delete[] vars;
	}

	glLinkProgram(handle);
	// check for link errors
	GLint succeeded;
	glGetProgramiv(handle, GL_LINK_STATUS, &succeeded);
	if (succeeded == GL_FALSE) {
		GLint logSize;
		glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);
		GLchar* log = new char[logSize];
		glGetProgramInfoLog(handle, logSize, nullptr, log);

		FATAL_ERROR("ERROR linking shader program: " << log);

		//delete[] log;	// never reached
	}

	// validate program (only in DEBUG)
#if _DEBUG
	glValidateProgram(handle);
	glGetProgramiv(handle, GL_VALIDATE_STATUS, &succeeded);
	if (succeeded == GL_FALSE) {
		GLint logSize;
		glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);
		GLchar* log = new char[logSize];
		glGetProgramInfoLog(handle, logSize, nullptr, log);

		FATAL_ERROR("ERROR validating shader program: " << log);

		//delete[] log;	// never reached
	}
#endif


	// don't need the shader handles any more
	for (unsigned int i = 0; i < shaderHandles.size(); i++)
		glDeleteShader(shaderHandles[i]);

	return handle;
}

GLint Shader::getLoc(std::string uniformName) {
	std::map<std::string, GLint>::iterator it;
	it = m_uniformLocations.find(uniformName);
	if (it == m_uniformLocations.end()) {
		// locate it with OpenGL and cache result
		GLint loc = glGetUniformLocation(m_handle, uniformName.c_str());
		m_uniformLocations[uniformName] = loc;
		if (loc == -1 && SHADER_OUTPUT_UNIFORM_WARNINGS) {
			std::cout << "WARNING: uniform \"" << uniformName << "\" not found in program " << m_name << std::endl;
		}
		return loc;
	}
	return it->second;
}

