#pragma once

#include <glm/gtc/type_ptr.hpp>
#include <map>
#include <vector>
#include "Utils.h"

const bool SHADER_OUTPUT_UNIFORM_WARNINGS = true; // output warnings for unfound uniforms?

//#define SHADER_PREVENT_BINDING_BOUND_SHADER	// is actually a tiny bit slower

#define SHADER_KEYWORD_NEEDS_SHADOWMAP  "//##SHADERLOADER:SHADER_NEEDS_SHADOWMAP##"
#define SHADER_KEYWORD_INSERTDEFINES    "//##SHADERLOADER:INSERTDEFINES##"

/// Represents a GLSL shader.
class Shader {
public:
	/// Simplified constructor.
	/// @see Shader(std::string, std::vector<std::string>, std::vector<std::string>, bool) with no tfbVaryings for details.
	Shader(std::string name, std::vector<std::string> filenames, bool usesMVP = false);
	/// Constructor.
	/// @param name        Name of the shader
	/// @param filenames   Shader filenames, relative to "shaders/" in the global assets directory
	/// @param tfbVaryings List of transform feedback varyings used in the shader (empty when not used with transform feedback)
	/// @param usesMVP     User-flag, stored in m_usesMVP. Not used by class Shader itself.
	Shader(std::string name, std::vector<std::string> filenames, std::vector<std::string> tfbVaryings, bool usesMVP = false);
	~Shader();

	inline std::string getName() { return m_name; }	///< Get shader name. @returns name
	inline GLuint getHandle() { return m_handle; }	///< Get shader handle. @returns handle

	/// Activate shader (set as current OpenGL shader).
	inline void use() {
#ifdef SHADER_PREVENT_BINDING_BOUND_SHADER
		if (this != Shader::m_CurrentlyBoundShader) {
			glUseProgram(m_handle);
			Shader::m_CurrentlyBoundShader = this;
		}
#else
		glUseProgram(m_handle);
#endif
	}

	/// Deactivate shader.
	static inline void unbind() { 
		glUseProgram(0);
#ifdef SHADER_PREVENT_BINDING_BOUND_SHADER
		m_CurrentlyBoundShader = 0;
#endif
	}

	/// Find Shader object with a given name.
	/// @param name Name of shader to find
	/// @returns Shader object
	/// @note If the shader is not found, a fatal error is triggered.
	static Shader * findShader(std::string name);
	static void deleteAllShaders();	///< Delete all loaded shaders.
	static std::vector<Shader *> & getAllShaders() { return m_AllShaders; }	///< Get a list of all loaded shaders. @returns vector of loaded Shader objects

	// Uniform functions
	GLint getLoc(std::string uniformName);	///< Get uniform location. @param uniformName Name of the uniform @returns location number (-1 if uniform is not found)
	inline void setInt(std::string name, GLint i)    { glUniform1i(getLoc(name), i); }											///< Set integer uniform. @param name Uniform name @param i integer value
	inline void setBool(std::string name, bool b)    { glUniform1i(getLoc(name), b ? 1 : 0); }									///< Set bool uniform.    @param name Uniform name @param b bool value
	inline void setFloat(std::string name, GLfloat f) { glUniform1f(getLoc(name), f); }											///< Set float uniform.   @param name Uniform name @param f float value
	inline void setVec2(std::string name, glm::vec2 v) { glUniform2fv(getLoc(name), 1, glm::value_ptr(v)); }					///< Set vec2 uniform.    @param name Uniform name @param v vec2 value
	inline void setIVec2(std::string name, glm::ivec2 v) { glUniform2iv(getLoc(name), 1, glm::value_ptr(v)); }					///< Set ivec2 uniform.   @param name Uniform name @param v ivec2 value
	inline void setVec3(std::string name, glm::vec3 v) { glUniform3fv(getLoc(name), 1, glm::value_ptr(v)); }					///< Set vec3 uniform.    @param name Uniform name @param v vec3 value
	inline void setVec4(std::string name, glm::vec4 v) { glUniform4fv(getLoc(name), 1, glm::value_ptr(v)); }					///< Set vec4 uniform.    @param name Uniform name @param v vec4 value
	inline void setMat4(std::string name, glm::mat4 m) { glUniformMatrix4fv(getLoc(name), 1, GL_FALSE, glm::value_ptr(m)); }	///< Set mat4 uniform.    @param name Uniform name @param m mat4 value
	inline void setMat3(std::string name, glm::mat3 m) { glUniformMatrix3fv(getLoc(name), 1, GL_FALSE, glm::value_ptr(m)); }	///< Set mat3 uniform.    @param name Uniform name @param m mat3 value
	// ... extend when needed

	bool m_usesMVP = false;   // premultiply projView with model matrix ?
	bool m_needsTex0 = false; // requires a texture

	static std::vector<std::string> m_customDefines; // inserted when shader code contains line: //##SHADERLOADER:INSERTDEFINES##

	inline bool getNeedsShadowmap() { return m_needsShadowmap; }

private:
	std::string loadShaderFile(std::string fileName);
	GLuint loadAndCompileShader(std::string fileName, GLenum shaderType);
	GLuint createShaderProgram(std::vector<std::string> filenames, std::vector<std::string> tfbVaryings);

	static std::vector<Shader *> m_AllShaders;
#ifdef SHADER_PREVENT_BINDING_BOUND_SHADER
	static Shader *m_CurrentlyBoundShader;
#endif

	std::string m_name;
	GLuint      m_handle;
	std::map<std::string, GLint> m_uniformLocations;

	bool m_needsShadowmap = false;
};
