#include <vector>
#include <Windows.h>
#include "Utils.h"

// Assets directory
std::string gAssetsDir = "assets";

/// Initialise global path to assets directory.
void initGlobals() {
	// find path to assets directory
	struct stat info;
	gAssetsDir = "assets";
	if (stat(gAssetsDir.c_str(), &info) != 0) {
		gAssetsDir = "../assets";
		if (stat(gAssetsDir.c_str(), &info) != 0) {
			gAssetsDir = "../../assets";
			if (stat(gAssetsDir.c_str(), &info) != 0) {
				gAssetsDir = "../../../assets";
				if (stat(gAssetsDir.c_str(), &info) != 0) {
					FATAL_ERROR("Cannot locate assets directory!")
				}
			}
		}
	}
	//std::cout << "Assets dir is \"" << gAssetsDir << "\"" << std::endl;

	// init renderdoc api
#if USE_RENDERDOC_API
	if(HMODULE mod = GetModuleHandleA("renderdoc.dll")) {
		pRENDERDOC_GetAPI RENDERDOC_GetAPI = (pRENDERDOC_GetAPI)GetProcAddress(mod, "RENDERDOC_GetAPI");
		int ret = RENDERDOC_GetAPI(eRENDERDOC_API_Version_1_1_2, (void **)&rdoc_api);
		assert(ret == 1);
		std::cout << "RenderDoc found, API enabled" << std::endl;
	}
#endif
}

struct FramebufferInfo;
std::vector<FramebufferInfo *> framebuffers;

/// Holds information to restore a framebuffer
struct FramebufferInfo {
	GLint draw;				///< Handle to OpenGL's draw-framebuffer object
	GLint read;				///< Handle to OpenGL's read-framebuffer object
	GLint viewport[4];		///< Viewport dimensions
	GLfloat clearcolor[4];	///< Clear colour
	//FramebufferInfo() { std::cout << "FBI created, cnt = " << framebuffers.size() << std::endl; }
	//~FramebufferInfo() { std::cout << "FBI destroyed, cnt = " << framebuffers.size() << std::endl; }
};

/// Store the current framebuffer and context
void getFramebufferInfo(FramebufferInfo &old) {
	glGetIntegerv(GL_DRAW_FRAMEBUFFER_BINDING, &old.draw);
	glGetIntegerv(GL_READ_FRAMEBUFFER_BINDING, &old.read);
	glGetIntegerv(GL_VIEWPORT, old.viewport);
	glGetFloatv(GL_COLOR_CLEAR_VALUE, old.clearcolor);
}

/// Restore a stored framebuffer and context
void restoreFramebuffer(const FramebufferInfo &old) {
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, old.draw);
	glBindFramebuffer(GL_READ_FRAMEBUFFER, old.read);
	glViewport(old.viewport[0], old.viewport[1], old.viewport[2], old.viewport[3]);
	glClearColor(old.clearcolor[0], old.clearcolor[1], old.clearcolor[2], old.clearcolor[3]);
}

/// Push the current framebuffer and context on a stack
void pushFramebuffer() {
	FramebufferInfo* inf = new FramebufferInfo();
	getFramebufferInfo(*inf);
	framebuffers.push_back(inf);
}

/// Pop framebuffer and context from stack
void popFramebuffer() {
	assert(!framebuffers.empty());
	if (!framebuffers.empty()) {
		FramebufferInfo* inf = framebuffers.back();
		framebuffers.pop_back();
		restoreFramebuffer(*inf);
		delete(inf);
	}
}



/// Format debug callback message from OpenGL
/// @note This function is taken from the framework provided for the ECG course 
std::string formatDebugOutput(GLenum source, GLenum type, GLuint id, GLenum severity, const char* msg) {
	std::stringstream stringStream;
	std::string sourceString;
	std::string typeString;
	std::string severityString;

	switch (source) {
	case GL_DEBUG_SOURCE_API: {
		sourceString = "API";
		break;
	}
	case GL_DEBUG_SOURCE_APPLICATION: {
		sourceString = "Application";
		break;
	}
	case GL_DEBUG_SOURCE_WINDOW_SYSTEM: {
		sourceString = "Window System";
		break;
	}
	case GL_DEBUG_SOURCE_SHADER_COMPILER: {
		sourceString = "Shader Compiler";
		break;
	}
	case GL_DEBUG_SOURCE_THIRD_PARTY: {
		sourceString = "Third Party";
		break;
	}
	case GL_DEBUG_SOURCE_OTHER: {
		sourceString = "Other";
		break;
	}
	default: {
		sourceString = "Unknown";
		break;
	}
	}

	switch (type) {
	case GL_DEBUG_TYPE_ERROR: {
		typeString = "Error";
		break;
	}
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: {
		typeString = "Deprecated Behavior";
		break;
	}
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: {
		typeString = "Undefined Behavior";
		break;
	}
	case GL_DEBUG_TYPE_PORTABILITY_ARB: {
		typeString = "Portability";
		break;
	}
	case GL_DEBUG_TYPE_PERFORMANCE: {
		typeString = "Performance";
		break;
	}
	case GL_DEBUG_TYPE_OTHER: {
		typeString = "Other";
		break;
	}
	case GL_DEBUG_TYPE_PUSH_GROUP: {
		typeString = "PushDebugGroup";
		break;
	}
	case GL_DEBUG_TYPE_POP_GROUP: {
		typeString = "PopDebugGroup";
		break;
	}
	default: {
		typeString = "Unknown (" + std::to_string(type) + ")";
		break;
	}
	}

	switch (severity) {
	case GL_DEBUG_SEVERITY_HIGH: {
		severityString = "High";
		break;
	}
	case GL_DEBUG_SEVERITY_MEDIUM: {
		severityString = "Medium";
		break;
	}
	case GL_DEBUG_SEVERITY_LOW: {
		severityString = "Low";
		break;
	}
	case  GL_DEBUG_SEVERITY_NOTIFICATION: {
		severityString = "Notification";
		break;
	}
	default: {
		severityString = "Unknown (" + std::to_string(severity) + ")";
		break;
	}
	}

	stringStream << "OpenGL Error: " << msg;
	stringStream << " [Source = " << sourceString;
	stringStream << ", Type = " << typeString;
	stringStream << ", Severity = " << severityString;
	stringStream << ", ID = " << id << "]";

	return stringStream.str();
}

/// Callback function for OpenGL error/warning messages.
void APIENTRY openGL_debugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const GLvoid* userParam) {
	if (id == 131185 || id == 131218 /* || (source == GL_DEBUG_SOURCE_APPLICATION && id == 0) */) return;
	//if (id == 131186 && gIgnoreTFBPerformanceWarning) return;

	std::string error = formatDebugOutput(source, type, id, severity, message);
	std::cout << error << std::endl;
}

RENDERDOC_API_1_1_2 *rdoc_api = NULL;
