/// @file meshmath.cpp
/// This file contains functions to calculate mesh properties.
///

//#define _CRT_SECURE_NO_WARNINGS
//#define _CRT_NONSTDC_NO_WARNINGS

/*
TODO:
Skel. fails with big meshes (like tree.obj) - why? 
CX8 still problematic
*/

#include "meshmath.h"
#include "Utils.h"
#include <iostream>
#include <igl/read_triangle_mesh.h>
//#include <igl/opengl/glfw/Viewer.h>
#include <igl/avg_edge_length.h>
#include <igl/heat_geodesics.h>
#include <igl/remove_duplicate_vertices.h>
#include <igl/cotmatrix.h>
#include <igl/massmatrix.h>
#include <igl/invert_diag.h>
//#include <igl/min_quad_with_fixed.h>
#include <igl/cat.h>
#include <igl/principal_curvature.h>
#include <igl/per_face_normals.h>
#include <igl/grad.h>
#include <igl/doublearea.h>
#include <igl/isolines_map.h>

#include <unsupported/Eigen/SparseExtra>	// for saving matrices

#include <math.h>


typedef Eigen::SparseMatrix<double> SpMat;

// ignore in doxygen, this is just debugging stuff
///@cond IGNORE
template <typename T>
void dump_matrix(int minRow, int maxRow, int minCol, int maxCol, const char *label, T &M, const char *fmt) {
	maxRow = std::min((long long)maxRow, M.rows() - 1);
	maxCol = std::min((long long)maxCol, M.cols() - 1);

	if (label) printf("%s (%d x %d):\n", label, M.rows(), M.cols());
	for (int i = minRow; i <= maxRow; i++) {
		printf("[%2d] ", i);
		for (int j = minCol; j <= maxCol; j++) {
			printf(fmt, M.coeff(i, j));
		}
		printf("\n");
	}
	printf("\n");
}

template <typename T>
void dump_matrix(const char *label, T &M, const char *fmt) {
	int maxi = std::min((long long)10, M.rows() - 1);
	int maxj = std::min((long long)10, M.cols() - 1);
	dump_matrix(0, maxi, 0, maxj, label, M, fmt);
}
///@endcond

/*
void calc_cotan_matrix(Eigen::MatrixXd &V, Eigen::MatrixXi &F, Eigen::MatrixXd &L) {
	// calculate nxn curvature flow Laplace operator:
	// def: w_ij = cot alpha_ij + cot beta_ij		(if edge(i,j), else 0)
	// L_ij = w_ij			if edge(i,j)
	//      = -sum(w_ik)	for all edge(i,k)
	//      = 0				else

	// TODO: can heavily optimize this if necessary

	int numV = V.rows();
	int numT = F.rows();

	L = Eigen::MatrixXd::Zero(numV, numV);

	for (int iTri = 0; iTri < numT; iTri++) {
		int i = F(iTri, 0); // A
		int j = F(iTri, 1); // B
		int k = F(iTri, 2); // C

		// cosine rule (law of cosines):
		//   c^2 = a^2 + b^2 - 2*a*b*cos(gamma)
		//	 gamma = acos(a^2 + b^2 - c^2 / 2ab)
		double a = (V.row(k) - V.row(j)).norm() ; // C-B
		double b = (V.row(i) - V.row(k)).norm() ; // A-C
		double c = (V.row(j) - V.row(i)).norm() ; // B-A

		double alpha = acos((b*b + c*c - a*a) / (2*b*c)); // edge(j,k)
		double beta  = acos((a*a + c*c - b*b) / (2*a*c)); // edge(i,k)
		double gamma = acos((a*a + b*b - c*c) / (2*a*b)); // edge(i,j)

		L(i,j) = 1.0 / tan(gamma);
		L(j,k) = 1.0 / tan(alpha);
		L(k,i) = 1.0 / tan(beta);
	}

	L = L + L.transpose().eval(); // add angles from opposite triangles	// NOTE: .eval() is necessary, otherwise the calc. will NOT work as expected!!!!

	// fill diagonal entries
	for (int iTri = 0; iTri < numT; iTri++) {
		int i = F(iTri, 0); // A
		int j = F(iTri, 1); // B
		int k = F(iTri, 2); // C
		L(i,i) = L(i,i) - (L(i,j) + L(i,k)) * 0.5; // half, as we process this twice
		L(j,j) = L(j,j) - (L(j,k) + L(j,i)) * 0.5;
		L(k,k) = L(k,k) - (L(k,i) + L(k,j)) * 0.5;
	}
}
*/

/// Calculate the (full) Cotangent Laplacian of a mesh
/// @param V Vertex matrix
/// @param F Face matrix
/// @param[out] L Cotangent Laplacian
void laplace_cotan_matrix(Eigen::MatrixXd& V, Eigen::MatrixXi& F, Eigen::SparseMatrix<double>& L) {
	igl::cotmatrix(V, F, L);
	L *= 2.0; // Note: igl::cotmatrix calcs half the Laplacian
}

/// Calculate one-ring areas of a mesh
/// @param V Vertex matrix
/// @param F Face matrix
/// @param[out] A One-ring area for each vertex (numV x 1)
/// @param[out] avgTriangleArea Average area of a mesh triangle
void one_ring_areas(Eigen::MatrixXd& V, Eigen::MatrixXi& F, Eigen::MatrixXd& A, double &avgTriangleArea) {
	// TODO: make A a VectorXd ?
	int numV = V.rows();
	int numT = F.rows();

	A = Eigen::MatrixXd::Zero(numV, 1);
	avgTriangleArea = 0.0;

	for (int iTri = 0; iTri < numT; iTri++) {
		int i = F(iTri, 0); // A
		int j = F(iTri, 1); // B
		int k = F(iTri, 2); // C

		Eigen::Vector3d b = (V.row(i) - V.row(k)); // A-C
		//Eigen::Vector3d c = (V.row(j) - V.row(i)); // B-A
		Eigen::Vector3d a = (V.row(k) - V.row(j)); // C-B

		//double area = b.cross(c).norm() * 0.5;
		double area = b.cross(a).norm() * 0.5;
		A(i) += area;
		A(j) += area;
		A(k) += area;
		avgTriangleArea += area;

	}

	A /= 3.0; // TODO: really?
	avgTriangleArea /= (double)numT;
}

/// Calculate the volume of a mesh
/// @param V Vertex matrix
/// @param F Face matrix
/// @returns Volume
double volume(Eigen::MatrixXd& V, Eigen::MatrixXi& F) {
	// TODO: optimize?

	int numV = V.rows();
	int numT = F.rows();

	double sum = 0.0;

	for (int iTri = 0; iTri < numT; iTri++) {
		int i = F(iTri, 0); // A
		int j = F(iTri, 1); // B
		int k = F(iTri, 2); // C

		//Eigen::Vector3d ab = V.row(j) - V.row(i);
		//Eigen::Vector3d ac = V.row(k) - V.row(i);
		//Eigen::Vector3d n = ab.cross(ac);
		//sum += V.row(i).dot(n);

		Eigen::Vector3d A = V.row(i);
		Eigen::Vector3d B = V.row(j);
		Eigen::Vector3d C = V.row(k);
		sum += A.dot(B.cross(C));
	}
	sum /= 6.0;
	return abs(sum);
}

// Implementation of "Skeleton Extraction by Mesh Contraction" [Au, 2008]
// modifies V
/// Extract the skeleton of a mesh.
///
/// This is an implementation of "Skeleton Extraction by Mesh Contraction" [Au, 2008]
/// @param[in,out] V Vertex matrix
/// @param F Face matrix
/// @param contraction_factor Contraction factor
/// @param maxIter Maximum number of iterations
///
/// @warning The vertex matrix V is modified by this function.
void skeleton_extraction(Eigen::MatrixXd &V, Eigen::MatrixXi &F, double contraction_factor, int maxIter) {
	const char *fmt = "%9.6f ";
	const double sL = contraction_factor; // contraction factor

	int numV = V.rows();

	Eigen::MatrixXd A;
	Eigen::SparseMatrix<double> L_sp;
	double A_avg;
	//Eigen::MatrixXd L;
	//calc_cotan_matrix(V, F, L);
	laplace_cotan_matrix(V, F, L_sp);
	one_ring_areas(V, F, A, A_avg);
	//dump_matrix("L", L_sp);
	//dump_matrix("A", A);
	//printf("Avg. triangle area: %f\n", A_avg);	// TODO: different in paper -> uses mean(A), i.e. avg of one-ring areas

	// initial conditions
	//Eigen::MatrixXd WH = Eigen::MatrixXd::Identity(numV, numV);
	////Eigen::MatrixXd WL = Eigen::MatrixXd::Ones(numV, numV) * sqrt(A_avg) * 1e-3;
	//Eigen::MatrixXd WL = Eigen::MatrixXd::Identity(numV, numV) * sqrt(A.mean()) * 1e-3;

	Eigen::SparseMatrix<double> WH_sp(numV,numV), WL_sp(numV,numV);
	WH_sp.setIdentity();
	WL_sp.setIdentity();
	WL_sp *= sqrt(A.mean()) * 1e-3;

	Eigen::MatrixXd A0(A);

	//dump_matrix("WL", WL, "%.8f  ");

	//Eigen::SparseMatrix<double> WH_sp = WH.sparseView();
	//Eigen::SparseMatrix<double> WL_sp = WL.sparseView();
	Eigen::SparseMatrix<double> bZero(numV,3);
	bZero.setZero(); // necessary?
	Eigen::SparseMatrix<double> WH0_sp(WH_sp);
	double vol0 = volume(V, F);

	//dump_matrix("A0", A0,    fmt);

	Eigen::MatrixXd Vprev;

	Eigen::SparseMatrix<double> M, WLL, WHV, b;
	for (int iter = 0; iter < maxIter; iter++) {
		Vprev = V;

		printf("Skeletonizing iter %d: ", iter);
		//dump_matrix("WL", WL_sp, fmt);
		//dump_matrix("WH", WH_sp, fmt);
		//dump_matrix("A",  A,     fmt);
		printf("vol = %f\n", volume(V, F));

		// set up matrices:
		// M = [ WL*L ]     b = [ 0    ]	//	M: 2n x n, b: 2n x 3
		//     [ WH   ]         [ WH*V ]
		//igl::cat(1, (WL_sp * L_sp).eval(), WH_sp, M); // hangs in 2nd iter???
		WLL = WL_sp * L_sp;
		WHV = (WH_sp * V).sparseView();
		igl::cat(1, WLL, WH_sp, M);
		igl::cat(1, bZero, WHV, b);

		// Step 1: solve for Vnew:
		// [ WL*L ] * Vnew = [ 0    ]		//	2n x n * n x 3  = 2n x 3
		// [ WH   ]          [ WH*V ]
		// i.e.: M * Vnew = b

		//M.makeCompressed();
		//b.makeCompressed();

		////Eigen::SparseQR<Eigen::SparseMatrix<double>, Eigen::COLAMDOrdering<int>> solver;	// SLOW! unusable - but good result
		//Eigen::LeastSquaresConjugateGradient<Eigen::SparseMatrix<double>> solver;	// way faster than SparseQR! but BAD results
		//solver.compute(M);
		//V = solver.solve(b);

		// SimplicialLDLT - this works FAST and delivers GOOD results!!
		Eigen::SparseMatrix<double> MtM = M.transpose() * M;
		Eigen::SparseMatrix<double> MtB = M.transpose() * b;
		Eigen::SimplicialLDLT<Eigen::SparseMatrix<double>> solver(MtM);
		V = solver.solve(MtB);

		// Conjugate Gradient - same bad results as LeastSquaresConjugateGradient
		//Eigen::SparseMatrix<double> MtM = M.transpose() * M;
		//Eigen::SparseMatrix<double> MtB = M.transpose() * b;
		//Eigen::ConjugateGradient<Eigen::SparseMatrix<double>> solver(MtM);
		//V = solver.solve(MtB);

		//Eigen::SparseLU<Eigen::SparseMatrix<double>> solver;	// crashes...
		//solver.analyzePattern(M);
		//solver.factorize(M);
		//V = solver.solve(b);

		// dump_matrix("Vnew", V);

		if (solver.info() != Eigen::ComputationInfo::Success) {
			switch (solver.info()) {
			case Eigen::ComputationInfo::InvalidInput:
				printf("SOLVER: invalid input!\n");
				break;
			case Eigen::ComputationInfo::NoConvergence:
				printf("SOLVER: no convergence!\n");
				break;
			case Eigen::ComputationInfo::NumericalIssue:
				printf("SOLVER: numerical issue!\n");
				break;
			default:
				printf("SOLVER: unknown error!\n");
				break;
			}
			V = Vprev;
			break;
		}

		{
			double relErr = (MtM * V - MtB).norm() / MtB.norm();
			printf("Rel.error: %f\n", relErr);
			if (isnan(relErr) || relErr > 0.001) { // NOT in paper!
				printf("Rel.error too large, stopping\n");
				V = Vprev;
				break;
			}
		}

		double vol = volume(V, F);
		if ((vol / vol0) < 1e-8) {
			printf("vol = %f: stopping iterations and using previous point set\n");
			V = Vprev;
			break;
		}

		if (isnan(vol) ||  (vol / vol0) < 1e-6) {
			printf("vol = %f: stopping iterations\n");
			V = Vprev; // NOT in original paper! - use the previous point set
			break;
		}

		// Step 2 and 3: update weights, recompute Laplace operator
		// WL_new = sL * WL
		// WH_new_i = WH0_i * sqrt(A0_i/A_i)		, where A_i is the current one ring area for vertex i (A0_i and WH0_i are the original values)

		laplace_cotan_matrix(V, F, L_sp);
		one_ring_areas(V, F, A, A_avg);

		WL_sp *= sL;
		Eigen::MatrixXd A_factors = A0.cwiseQuotient(A).cwiseSqrt(); // element-wise division, sqrt
		Eigen::DiagonalMatrix<double, Eigen::Dynamic> diag = A_factors.asDiagonal();
		Eigen::SparseMatrix<double> f;
		f = diag;
		WH_sp = WH0_sp.cwiseProduct(f);

		// TODO: recheck A-decay!!

		//dump_matrix("A0", A0, "%7.4f ");
		//dump_matrix("A", A, "%7.4f ");
		//dump_matrix("A_factors", A_factors, "%7.4f ");
		//dump_matrix("f", f, "%7.4f ");
		//break;
	}

	//Eigen::SparseMatrix<double> sp = V.sparseView();
	//Eigen::saveMarket(sp, "D:/tmp/testmatrix.mtx");

}


// ignore in doxygen, this is just debugging stuff
///@cond IGNORE
void solver_test() {
	Eigen::MatrixXd Adense(4, 3), xxdense(3, 2), bdense;
	Adense << 1, 2, 3,
		5, 9, 7,
		3, 1, 4,
		8, 8, 2;
	xxdense << 3, 6,
		5, 10,
		7, 14;
	bdense = Adense * xxdense;
	for (int i = 0; i < bdense.rows(); i++) for (int j = 0; j < bdense.cols(); j++) bdense(i, j) += (double)(i + 1) * 0.01;

	Eigen::SparseMatrix<double> A = Adense.sparseView(), b = bdense.sparseView();

#if 0
	A.makeCompressed();
	Eigen::SparseQR<Eigen::SparseMatrix<double>, Eigen::COLAMDOrdering<int>> solverA;
	solverA.compute(A);

	// ok, but also works with b as matrix..
	//Eigen::VectorXd b0 = bdense.col(0), b1 = bdense.col(1);
	//Eigen::VectorXd x0 = solverA.solve(b0);
	//printf("x0 (%d x %d):\n", x0.rows(), x0.cols());
	//for (int i = 0; i < x0.rows(); i++) printf("%f\n", x0(i, 0));

	{
		Eigen::MatrixXd x = solverA.solve(bdense);
		printf("(1) x (%d x %d):\n", x.rows(), x.cols());
		Eigen::MatrixXd xMat = x; dump_matrix(nullptr, xMat, "%7.4f ");
	}
	{
		// b can be sparse too
		Eigen::MatrixXd x = solverA.solve(b);
		printf("(2) x (%d x %d):\n", x.rows(), x.cols());
		Eigen::MatrixXd xMat = x; dump_matrix(nullptr, xMat, "%7.4f ");
	}
#else
	// testing LeastSquaresConjugateGradient
	Eigen::LeastSquaresConjugateGradient<Eigen::SparseMatrix<double>> solver;
	solver.compute(A);
	Eigen::MatrixXd x = solver.solve(b);
	printf("(LSCG) x (%d x %d):\n", x.rows(), x.cols());
	Eigen::MatrixXd xMat = x; dump_matrix(nullptr, xMat, "%7.4f ");
#endif


	// TODO: test mqwf
	//igl::min_quad_with_fixed_data<double> mqwf;
	//Eigen::SparseMatrix<double> Aeq, beq; // no constraints

	//Eigen::VectorXd b = bdense.col(0);
	//igl::min_quad_with_fixed_precompute(A, b, Aeq, false, mqwf);
	//igl::min_quad_with_fixed_solve()


}
///@endcond

/// Find endpoints on a skeletonised mesh
///
/// This is an implementation of a part of "Parameterization and Feature Extraction for the Visualization of Tree-like Structures" [Lichtenberg 2018]
/// @param V Vertex matrix (of *skeleton* mesh)
/// @param F Face matrix
/// @param[out] ixEndpoints Vertex indices of located endpoints
void find_endpoints(Eigen::MatrixXd &V, Eigen::MatrixXi &F, std::vector<int> *ixEndpoints) {
	// V are skeletonized points

	int numV = V.rows();
	int numT = F.rows();

	// calc average edge length
	double avg_edge_length = 0.0;
	for (int iTri = 0; iTri < numT; iTri++) {
		Eigen::Vector3d A = V.row(F(iTri, 0));
		Eigen::Vector3d B = V.row(F(iTri, 1));
		Eigen::Vector3d C = V.row(F(iTri, 2));
		avg_edge_length += (B - A).norm() + (C - B).norm() + (A - C).norm();
	}
	avg_edge_length /= 3.0 * numT;

	double radius = 3.0 * avg_edge_length;

	std::vector<int> ix;
	ix.reserve(numV);

	for (int iV = 0; iV < numV; iV++) {
		Eigen::Vector3d center = V.row(iV);

		// distance from center point to all points
		Eigen::MatrixXd allCenter = center.transpose().colwise().replicate(numV);	// n x 3, each row is center
		Eigen::MatrixXd vecs = allCenter - V;
		Eigen::VectorXd dist = vecs.cwiseProduct(vecs).rowwise().sum().cwiseSqrt();

		// keep points with dist < radius
		// no equivalent to matlab find() in Eigen ??
		ix.clear();

		// find point inside radius with largest distance to center
		int    iFurthest = iV;
		double dFurthest = 0.0;
		for (int i = 0; i < numV; i++) {
			float d = dist(i);
			if (d < radius) {
				ix.push_back(i);
				if (d > dFurthest) {
					iFurthest = i;
					dFurthest = d;
				}
			}
		}

		//printf("%d: %d points in radius\n", iV, ix.size());

		if (iFurthest != iV) {
			Eigen::Vector3d n = V.row(iFurthest) - V.row(iV);
			// check if all of the points lie on the same half-plane
			bool ok = true;
			for (int i = 0; i < (int)ix.size(); i++) {
				Eigen::Vector3d v = V.row(ix[i]) - V.row(iV);
				if (n.dot(v) < 0.0) {
					ok = false;
					break;
				}
			}
			if (ok) {
				//printf("Found endpoint at ix %d\n", iV);
				if (ixEndpoints) ixEndpoints->push_back(iV);
			}
		}
	}
}

/// Find root point of a mesh
/// @param ixEndpoints Vertex indices of endpoints, as calculated by find_endpoints()
/// @param V Vertex matrix (of the *original* mesh, **not** the skeleton)
/// @param F Face matrix
/// @returns Vertex index of the root point
int find_root_point(std::vector<int> ixEndpoints, Eigen::MatrixXd &V, Eigen::MatrixXi &F) {
	// V is the original (not skeletonized) vertex set!

	if (ixEndpoints.size() == 0) return -1;

	// calculate mean curvature - NOTE: This code snipped is from IGL tutorial 203
	Eigen::MatrixXd HN;
	Eigen::SparseMatrix<double> L,M,Minv;
	igl::cotmatrix(V,F,L);
	igl::massmatrix(V,F,igl::MASSMATRIX_TYPE_VORONOI,M);
	igl::invert_diag(M,Minv);
	// Laplace-Beltrami of position
	HN = -Minv*(L*V);
	// Extract magnitude as mean curvature
	Eigen::VectorXd H = HN.rowwise().norm();


	// use endpoint where absolute mean curvature is minimal
	int iMin = ixEndpoints[0];
	double hMin = H(iMin);
	for (int i = 1; i < (int)ixEndpoints.size(); i++) {
		if (H(ixEndpoints[i]) < hMin) {
			iMin = ixEndpoints[i];
			hMin = H(iMin);
		}
	}

	printf("Root point index is %d\n", iMin);
	return iMin;
}

/// Remove duplicate vertices from a mesh.
/// @param[in,out] V Vertex matrix
/// @param[in,out] F Face matrix
/// @warning Both V and F are modified by this function.
// .. because method fails on meshes with duplicate vertices (e.g. CX8_..)
void cleanup_mesh(Eigen::MatrixXd& V, Eigen::MatrixXi& F) {
	// model cx8 - right renal vein fails on geodesic precomp -> numerical issue; => removing duplicate vertices with meshlab helped! -> following code helps too
	Eigen::MatrixXd SV;
	Eigen::MatrixXi SVI, SVJ, SF;
	igl::remove_duplicate_vertices(V, F, 1e-7, SV, SVI, SVJ, SF);
	if (V.rows() != SV.rows()) {
		printf("Removed duplicate vertices, new mesh has %d vertices and %d faces\n", SV.rows(), SF.rows());
		V = SV; F = SF;
	}
}

/// Calculate the principal curvatures of a mesh.
/// @param V Vertex matrix
/// @param F Face matrix
/// @param[out] PDmin Direction of the minimal principal curvature
/// @param[out] PDmax Direction of the maximal principal curvature
/// @param[out] PVmin Magnitude of the minimal principal curvature
/// @param[out] PVmax Magnitude of the maximal principal curvature
void curvature(Eigen::MatrixXd& V, Eigen::MatrixXi& F, Eigen::MatrixXd& PDmin, Eigen::MatrixXd& PDmax, Eigen::VectorXd& PVmin, Eigen::VectorXd& PVmax) {
	// NOTE: igl documentation seems wrong (min/max parameter orders are swapped for both PD and PV)
	igl::principal_curvature(V, F, PDmin, PDmax, PVmin, PVmax);
	return;
}

/*
void debug_show_curv(Eigen::MatrixXd& V, Eigen::MatrixXi& F, Eigen::MatrixXd *curvB, Eigen::MatrixXd *curvR = nullptr) {
	using namespace Eigen;

	if (curvB && curvB->rows() != V.rows()) FATAL_ERROR("invalid curvB");
	if (curvR && curvR->rows() != V.rows()) FATAL_ERROR("invalid curvR");

	igl::opengl::glfw::Viewer viewer;
	viewer.data().set_mesh(V, F);

	const double avg = igl::avg_edge_length(V,F);

	const RowVector3d red(0.8,0.2,0.2),blue(0.2,0.2,0.8);

	if (curvB) viewer.data().add_edges(V, V + (*curvB) * avg, blue);
	if (curvR) viewer.data().add_edges(V, V + (*curvR) * avg, red);
	viewer.data().add_points(V, blue);
	viewer.data().point_size = 5;

	// Hide wireframe
	viewer.data().show_lines = false;
	viewer.launch();
	return;
}
*/

/// Initialise a sparse matrix from a diagonal-vector.
///
/// The resulting sparse matrix is zero everywhere, except on its diagonal; the diagonal elements are filled in from the vector diagV.
/// @param diagV Vector containing the diagonal elements
/// @param[out] sp Resultant sparse matrix
void initSparseMatFromDiag(Eigen::VectorXd& diagV, Eigen::SparseMatrix<double> &sp) {
	// TODO: is this effiecient? or better use triplet initialization?
	Eigen::DiagonalMatrix<double,Eigen::Dynamic> dMat = diagV.asDiagonal();
	sp = dMat;
}

/// Perform isotropic smoothing on a matrix, given the Cotangent Laplacian of the mesh
/// @param[in,out] M Matrix to smooth (number of rows = number of mesh vertices)
/// @param numIterations Number of smoothing iterations
/// @param weight Smoothing weight
/// @param Laplacian Cotangent Laplacian of the mesh
/// @warning M is modified by this function.
void isotropic_smoothing(Eigen::MatrixXd &M, int numIterations, double weight, Eigen::SparseMatrix<double> &Laplacian) {
	// perform isotropic smoothing on M, given the Laplacian

	using namespace Eigen;

	SpMat L(0.5 * Laplacian);

	VectorXd rDiag = -1.0 / L.diagonal().array();
	SpMat rDiagMat;
	initSparseMatFromDiag(rDiag, rDiagMat);
	L = rDiagMat * L;

	MatrixXd tmp;
	for (int i = 0; i < numIterations; i++) {
		tmp = L * M * weight;
		M += tmp;
	}

}

/// Convert per-vertex curvature to per-triangle curvature
/// @param F Face matrix
/// @param FaceNormal Matrix of face normals (numF x 3)
/// @param vcurv Per-vertex curvature (numV x 3)
/// @param[out] tcurv Per-triangle curvature (numF x 3), **preallocated**
/// @attention The output matrix tcurv must already be of correct size.
void vertex_to_triangle_curvature(Eigen::MatrixXi &F, Eigen::MatrixX3d &FaceNormal, Eigen::MatrixXd &vcurv, Eigen::MatrixXd &tcurv) {
	// convert per-vertex curvature to per-triangle curvature
	// tcurv must already be of correct size (numT x 3)!

	// TODO: optimize if necessary, using matrix-ops

	using namespace Eigen;

	int numT = F.rows();
	if (tcurv.rows() != numT || tcurv.cols() != 3) FATAL_ERROR("tcurv is of wrong size");

	for (int iTri = 0; iTri < numT; iTri++) {
		VectorXd d = (vcurv.row(F(iTri, 0)) + vcurv.row(F(iTri, 1)) + vcurv.row(F(iTri, 2))) / 3.0; // average curv dir of vertices
		double n_dot_d = FaceNormal.row(iTri) * d; // subtract projection on face normal
		tcurv.row(iTri) = (d - FaceNormal.row(iTri).transpose() * n_dot_d).normalized().transpose();
	}
}

/// Convert per-triangle curvature to per-vertex curvature
/// @param F Face matrix
/// @param tcurv Per-triangle curvature (numF x 3)
/// @param[out] vcurv Per-vertex curvature (numV x 3), **preallocated**
/// @attention The output matrix vcurv must already be of correct size.
void triangle_to_vertex_curvature(Eigen::MatrixXi &F, Eigen::MatrixXd &tcurv, Eigen::MatrixXd &vcurv) {
	// convert per-triangle-curvature to per-vertex curvature
	// vcurv must already be of correct size (numV x 3)!

	using namespace Eigen;

	int numT = F.rows();

	vcurv.setZero();
	for (int iTri = 0; iTri < numT; iTri++) {
		for (int i = 0; i < 3; i++) {
			vcurv.row(F(iTri,i)) += tcurv.row(iTri);
		}
	}
	vcurv.rowwise().normalize();
}

/// Calculate the U-parameterisation of the mesh.
///
/// This is an implementation of a part of "Parameterization and Feature Extraction for the Visualization of Tree-like Structures" [Lichtenberg 2018]
/// @param V Vertex matrix
/// @param F Face matrix
/// @param rootPointIndex Index of the root point vertex
/// @param[out] U U-parameter (per vertex)
/// @param[out] pStorePVmin When specified, the minimal principal curvature magnitude of the mesh is stored there. Can be nullptr.
void calculate_U(Eigen::MatrixXd &V, Eigen::MatrixXi &F, int rootPointIndex, Eigen::VectorXd &U, Eigen::VectorXd *pStorePVmin) {
	using namespace Eigen;

	// TODO: check usage of half/full Laplacian in all steps, esp. grad, div!

	int numV = V.rows();
	int numT = F.rows();

	// calculate principal curvature per vertex
	MatrixXd PDmin, PDmax; // pr.curv. directions
	VectorXd PVmin, PVmax; // pr.curv. values
	curvature(V, F, PDmin, PDmax, PVmin, PVmax);

	if (pStorePVmin) *pStorePVmin = PVmin;

	// just for testing
	//{ SpMat tmp; Eigen::loadMarket(tmp, "D:/tmp/pcd1.mtx"); PDmin = tmp; }

	// calculate per-triangle normals
	MatrixX3d tri_normals;
	igl::per_face_normals(V, F, tri_normals);

	// calculate per-triangle min curvature
	MatrixXd tri_curv(numT, 3);
	vertex_to_triangle_curvature(F, tri_normals, PDmin, tri_curv);
	//for (int iTri = 0; iTri < numT; iTri++) {
	//	Vector3d d = (PDmin.row(F(iTri, 0)) + PDmin.row(F(iTri, 1)) + PDmin.row(F(iTri, 2))) / 3.0; // average curv dir of vertices
	//	// subtract projection on face normal
	//	double n_dot_d = tri_normals.row(iTri) * d;
	//	tri_curv.row(iTri) = (d - tri_normals.row(iTri).transpose() * n_dot_d).normalized().transpose();
	//}

	// compute geodesic distance of each vertex from root point (Geodesics in Heat method)
	igl::HeatGeodesicsData<double> heat_data;
	double heat_t = std::pow(igl::avg_edge_length(V,F),2);
	if (!igl::heat_geodesics_precompute(V, F, heat_t, heat_data)) FATAL_ERROR("Geodesics in Heat: precompute failed");
	Eigen::VectorXd heat_d; // distances
	igl::heat_geodesics_solve(heat_data, (Eigen::VectorXi(1, 1) << rootPointIndex).finished(), heat_d);

	// calc the gradient operator for the mesh
	SpMat Grad;
	igl::grad(V, F, Grad);

	// calc gradient of geodesic distances
	MatrixXd gradG = Map<const MatrixXd>((Grad * heat_d).eval().data(),F.rows(),3);
	gradG.rowwise().normalize(); // necessary?

	// fix curvature directions via gradG
	for (int iTri = 0; iTri < numT; iTri++) {
		if (tri_curv.row(iTri) * gradG.row(iTri).transpose() < 0.0) {
			tri_curv.row(iTri) *= -1.0;
		}
	}

	// build new per-vertex curvature from per-triangle curv.
	MatrixXd vert_curv(numV, 3);
	triangle_to_vertex_curvature(F, tri_curv, vert_curv);
	//vert_curv.setZero();
	//for (int iTri = 0; iTri < numT; iTri++) {
	//	for (int i = 0; i < 3; i++) {
	//		vert_curv.row(F(iTri,i)) += tri_curv.row(iTri);
	//	}
	//}
	//vert_curv.rowwise().normalize();

	// smooth the new vertex-curvature a bit
	//{ SpMat tmp; Eigen::loadMarket(tmp, "D:/tmp/Xv.mtx"); vert_curv = tmp; }

	SpMat Laplace;
	laplace_cotan_matrix(V, F, Laplace);

	// base # smoothing iterations on root diameter and avg. edge length
	double diaRoot = 2.0 / (abs(PVmin(rootPointIndex)) + abs(PVmax(rootPointIndex)));
	int numIter = int(30.0 * diaRoot / igl::avg_edge_length(V, F));
	printf("# smoothing iter: %d\n", numIter);

	isotropic_smoothing(vert_curv, numIter, 0.5, Laplace);

	// and convert it back to per-triangle-curvature again
	vertex_to_triangle_curvature(F, tri_normals, vert_curv, tri_curv);

	//{ SpMat tmp; Eigen::loadMarket(tmp, "D:/tmp/X_smoothed.mtx"); tri_curv = tmp; }



	// compute discrete divergence operator:
	// build "triplicated" per-triangle mass matrix:
	//  M = [ A/3 0   0  ]
	//      [ 0   A/3 0  ]
	//      [ 0   0   A/3]
	// then obtain Div = Grad.transpose() * M
	VectorXd Athird;
	igl::doublearea(V, F, Athird);	// 2 * area of triangles
	Athird = Athird.colwise().replicate(3).eval() / 6.0;
	SpMat tri_mass3;
	initSparseMatFromDiag(Athird, tri_mass3);
	SpMat Div;
	Div = Grad.transpose() * tri_mass3;
	//dump_matrix("Div", Div);


	// reshape tri_curv to match div operator: Z = [x0 x1 ... x_, y0 y1 ... y_, z0 z1 ... z_]'   ->   (3*numT x 1) vector
	Map<VectorXd> Z(tri_curv.data(), tri_curv.size()); // Note: tri_curv must be stored in column major format! (Eigen default)

	// solve Poisson equation:
	// Grad U = Z  ===> Div . Grad U = Div. Z ===> Laplacian U = Div . Z ===> U = Laplacian \ (Div . Z)
	// Ax = b (A=L, b=Div.Z)   :  (nv x nv) * (nv x 1) = (nv x 1)

	// TODO: test other solvers (L is negative semi-definite, so (-L) is positive semi-definite, may exploit that fact...)
	SpMat& A = Laplace;
	VectorXd b;
	b = Div * Z;

	//SpMat AtA = A.transpose() * A;
	//MatrixXd AtB = A.transpose() * b;
	//Eigen::SimplicialLDLT<Eigen::SparseMatrix<double>> solver(AtA);
	//U = solver.solve(AtB);

	Eigen::SimplicialLDLT<Eigen::SparseMatrix<double>> solver(A);
	U = solver.solve(b);

	if (solver.info() != Eigen::ComputationInfo::Success) {
		switch (solver.info()) {
		case Eigen::ComputationInfo::InvalidInput:
			printf("SOLVER: invalid input!\n");
			break;
		case Eigen::ComputationInfo::NoConvergence:
			printf("SOLVER: no convergence!\n");
			break;
		case Eigen::ComputationInfo::NumericalIssue:
			printf("SOLVER: numerical issue!\n");
			break;
		default:
			printf("SOLVER: unknown error!\n");
			break;
		}
		FATAL_ERROR("Solver failed");
	}

	//dump_matrix("A", A);
	//dump_matrix("b", b);
	//dump_matrix("U", U);

	//double relErr = (A * U - b).norm() / b.norm();
	//printf("Rel.err %f\n", relErr);


	// map U to range of geodesic distances
	double minU = U.minCoeff();
	double maxU = U.maxCoeff();
	double maxG = heat_d.maxCoeff();
	U = (maxG * (1.0 - (U.array() - minU) / (maxU - minU))).matrix().eval();

	//printf("minG=%f maxG=%f\n", heat_d.minCoeff(), maxG);

	return;


	//{
	//	SpMat tmp; Eigen::loadMarket(tmp, "D:/tmp/Xv.mtx"); MatrixXd ref = tmp;
	//	debug_show_curv(V, F, &vert_curv, &ref); return;
	//	// TODO: generally ok, a few noticable differences...
	//}


	//debug_show_curv(V, F, &vert_curv, &PDmin); return;



	//dump_matrix("PD1", PD1);
	//dump_matrix("PD2", PD2);
	//dump_matrix("PV1", PV1);
	//dump_matrix("PV2", PV2);

}


