#version 430 core

// Basic idea inspired by Section 4.3 of:
// 
// Balzs Hajagos, Lszl Szcsi, and Balzs Csbfalvi. 2012.
// Fast silhouette and crease edge synthesis with geometry shaders.
// In Proceedings of the 28th Spring Conference on Computer Graphics (SCCG 12). Association for Computing Machinery, New York, NY, USA, 7176.
// DOI:https://doi.org/10.1145/2448531.2448540
//
// we don't use adjacent triangle information though

// Need to either disable backface culling or use flipping..

layout (triangles) in;
layout (triangle_strip, max_vertices = 7) out;

in VertexData {
	vec3 positionVS;
	vec3 normalVS;
	vec2 UV;
} gs_in[3];

out GeoOut {
	vec3 positionVS;
	vec3 normalVS;
	float paramU; // TODO!!
	float paramQ;
	float isContour;
} gs_out;

uniform float contourSize = 0.02;
uniform bool doContour = true;
uniform bool orthoProjection = false;
uniform mat4 matP;

vec3 alpha; // alpha[i] contains dot(normal[i], camDir[i])

// to find intersection point of contour along edge i1/i2:
// res = (alpha[i1] * p[i2] - alpha[i2] * p[i1]) / (alpha[i1] - alpha[i2]);
//
// note: this is the same as linearly interpolating with:
// f = -alpha[i2] / (alpha[i1] - alpha[i2])
// res = p[i1] * f + p[i2] * (1-f) ==  mix(p[i2], p[i1], f)

float contourFactor(int i1, int i2) {
	return -alpha[i2] / (alpha[i1] - alpha[i2]);
}

vec3 toContour(vec3 vA, vec3 vB, float mixFactor) {
	return mix(vB, vA, mixFactor);
}

float toContour(float fA, float fB, float mixFactor) {
	return mix(fB, fA, mixFactor);
}

void outputVertex(vec3 posVS, vec3 nrmVS, float u, float q, float isContour) {
	gl_Position = matP * vec4(posVS, 1);
	gs_out.positionVS  = posVS;
	gs_out.normalVS    = nrmVS;
	gs_out.paramU      = u;
	gs_out.paramQ      = q;
	gs_out.isContour   = isContour;
	EmitVertex();
}

void makeQuad(int i1, int i2, int i3, bool flip) {
	// edges i1-i3 and i2-i3 intersect contour (i1 and i2 are on the "same side", i.e. both either visible or invisible)
	float f13 = contourFactor(i1, i3);
	vec3  p13 = toContour(gs_in[i1].positionVS, gs_in[i3].positionVS, f13);
	vec3  n13 = toContour(gs_in[i1].normalVS,   gs_in[i3].normalVS,   f13);
	float f23 = contourFactor(i2, i3);
	vec3  p23 = toContour(gs_in[i2].positionVS, gs_in[i3].positionVS, f23);
	vec3  n23 = toContour(gs_in[i2].normalVS,   gs_in[i3].normalVS,   f23);

	// draw quad parallel to screen
	//n13.z = 0; n23.z = 0;
	n13 = normalize(vec3(n13.xy,0));
	n23 = normalize(vec3(n23.xy,0));

	vec3 pA = p13;
	vec3 pB = p23;
	vec3 pC = p23 + n23 * contourSize;
	vec3 pD = p13 + n13 * contourSize;

	// u,v for contours
	float u13 = toContour(gs_in[i1].UV.x, gs_in[i3].UV.x, f13);
	float u23 = toContour(gs_in[i2].UV.x, gs_in[i3].UV.x, f23);
	float qi  = 0.0; // inner v
	float qo  = 1.0; // outer v


	outputVertex(pA, n13, u13, qi, 1);
	if (!flip) {
		outputVertex(pB, n23, u23, qi, 1);
		outputVertex(pD, n13, u13, qo, 1);
	} else {
		outputVertex(pD, n13, u13, qo, 1);
		outputVertex(pB, n23, u23, qi, 1);
	}
	outputVertex(pC, n23, u23, qo, 1); 
	EndPrimitive();
}

void main() {

	if (doContour) {
		// dot vertex normal with to-eye-vector
		if (orthoProjection) {
			vec3 toEye = vec3(0,0,1);
			alpha[0] = dot(gs_in[0].normalVS, toEye);
			alpha[1] = dot(gs_in[1].normalVS, toEye);
			alpha[2] = dot(gs_in[2].normalVS, toEye);
		} else {
			alpha[0] = dot(gs_in[0].normalVS, -gs_in[0].positionVS);
			alpha[1] = dot(gs_in[1].normalVS, -gs_in[1].positionVS);
			alpha[2] = dot(gs_in[2].normalVS, -gs_in[2].positionVS);
		}

		bvec3 vis = greaterThanEqual(alpha, vec3(0));

		if (!all(vis) && !all(!vis)) {
			gs_out.isContour = 1.0;

			if      (vis.x == vis.y) makeQuad(0,1,2,vis.z);
			else if (vis.y == vis.z) makeQuad(1,2,0,vis.x);
			else if (vis.z == vis.x) makeQuad(2,0,1,vis.y);
		}
	}

	// output the triangle
	gs_out.isContour = 0.0;
	for (int i = 0; i < 3; i++) {
		outputVertex(gs_in[i].positionVS, gs_in[i].normalVS, gs_in[i].UV.x, gs_in[i].UV.y, 0);
	}
	EndPrimitive();

}
