#version 430 core

layout (location=0) out vec4 fragColor;

uniform vec3 dirLight_direction_VS = vec3(0,0,-1); // normalized!
uniform float ambientLight = 0.1;

uniform float mat_diffuse  = 0.7;
uniform float mat_specular = 0.3;
uniform float mat_shiny = 8.0;

uniform float u_scale = 1.0;
//uniform int uv_mode = 0;
uniform vec3 bgColor = vec3(0,1,1);
uniform vec3 contourColor = vec3(0,1,0);
uniform bool hasContours = false;
uniform ivec2 screenSize;
uniform float time = 0.0;
uniform float debug_float = 0.0;

uniform vec3 mesh_base_color   = vec3(1,1,0);	// if no color gradient is used

uniform bool use_blinn_phong = true;

uniform struct {
	bool enabled;
	bool onContour;
	bool useTexture;
	bool invert;
} scalar_field;

uniform struct {
	bool enabled;
	float f,o,w,l,s;
	int n;
} hatching;

uniform struct {
	bool  enabled;
	float freq;
	float width;
} isolines_u, isolines_v;

uniform struct {
	bool enabled;
	bool onContour;
	vec3 scale; // .z is used instead of .y for contour-decals
	vec2 offset;
	bool clampV, invertV, animate;
} decals;

uniform sampler2D texPositionVS;
uniform sampler2D texNormalVS;		// .xyz normal, .w: u-param
uniform sampler2D texParamV;
uniform sampler2D texDecal;
uniform sampler2D texGrad;

const float PI = 3.1415926535897932384626433832795;
const float TWO_PI = 2.0 * PI;

const float additional_v_scale = 2.0f; // empirical

//float rand() {
//	vec2 c = gl_FragCoord.xy / screenSize;
//    return fract(sin(dot(c.xy ,vec2(12.9898,78.233))) * 43758.5453);
//}

float rand1(float seed) {
	return fract(sin(seed) * 43758.5453123);
}
float rand1_range(float seed, vec2 range) {
	return rand1(seed) * (range.y - range.x) + range.x;
}
float rand2(vec2 seed) {
	return fract(sin(dot(seed, vec2(12.9898, 4.1414))) * 43758.5453);
}
float rand2_range(vec2 seed, vec2 range) {
	return rand2(seed) * (range.y - range.x) + range.x;
}

vec3 scalar_field_color(float uNormed) {
	float u = scalar_field.invert ? 1.0 - uNormed : uNormed;
	if (!scalar_field.useTexture) return mix(vec3(0,0,1), vec3(0,1,0), u);
	return texture(texGrad, vec2(u,0)).rgb;
}

vec3 calc_base_color(float uNormed) {
	if (scalar_field.enabled && !scalar_field.onContour) return scalar_field_color(uNormed);
	return mesh_base_color;
}

float calc_isolines_factor(float uNormed, float vNormed) {
	// return a factor to scale the base color by
	bool uu = isolines_u.enabled && (sin(uNormed * isolines_u.freq * TWO_PI)) * 0.5 + 0.5 > 1.0 - isolines_u.width * isolines_u.freq * 0.001;
	bool vv = isolines_v.enabled && (sin(vNormed * isolines_v.freq * TWO_PI)) * 0.5 + 0.5 > 1.0 - isolines_v.width * isolines_v.freq * 0.001;
	return !uu && !vv ? 1.0 : 0.0;
}

float calc_hatching_factor(float uNormed, float vScaled) {
		if (!hatching.enabled) return 1.0;

		float u = uNormed;
		float v = vScaled;

		int numHatches = hatching.n; // 5;
		float hFinal = 1;
		for (int i = 0; i < numHatches; i++) {
			float o = hatching.o;

			if (numHatches > 1) {
				//o = i * 0.1;
				//o += v * 0.001 * rand1_range((i+1)*u, vec2(-1,1)) + float(i) * debug_float;//0.005;
				o += v * 0.001 * (float(i)-(numHatches-1)/2)/(numHatches-1) + float(i) * debug_float;//0.005;
				//o = v * debug_float;	// +- 0.003 is a good range
				//o += v * debug_float * (float(i)-(numHatches-1)/2)/(numHatches-1);
			}

			float val = (u + o) * hatching.f;
			float b = floor(val);
			float u_hat = val - b;

			float l = hatching.l;
			float w = hatching.w;

			l *= rand2_range(vec2(b,2*i  ), vec2(0.2, 1));
			//w *= rand2_range(vec2(b,2*i+1), vec2(0.8, 1));

			//float h = cos(u_hat * 2 * PI) * 0.5 + 0.5 + w;
			//float h = clamp(cos(u_hat * 2 * PI) * 0.5 + 1.5 - 2.0 * w, 0, 1);
			float h = clamp(pow(cos(u_hat * 2 * PI) * 0.5 + 0.5, w), 0, 1);


			float t = clamp(pow(v / l, hatching.s), 0, 1);
			h = (1 - t) * h + t;

			//hFinal *= h;
			hFinal = min(hFinal, h);
		}

		return hFinal;
}

vec4 getContourColor(vec2 uv) {
	vec4 color = vec4(contourColor,1);

	// contour-scalar field
	if (scalar_field.enabled && scalar_field.onContour) {
		color.rgb = scalar_field_color(uv.x);
	}

	// contour-decals
	if (decals.enabled && decals.onContour) {
		float u = uv.x * 20;
		float v = uv.y;
		float t = decals.animate ? time : 1.0;

		vec2 texUV = vec2(u,v) * decals.scale.xz + decals.offset * vec2(t,1);
		if (decals.clampV)  texUV.y = clamp(texUV.y, 0, 1);
		if (decals.invertV) texUV.y = -texUV.y;

		vec4 colTex = texture(texDecal, texUV);
		if (decals.clampV && (abs(texUV.y) >= 1.0 || abs(texUV.y) <= 0.0)) colTex = vec4(0); // mipmapping will leave artefacts otherwise
		//color = mix(color, colTex.rgb, colTex.a);
		//return vec4(colTex.rgb,0.5);
		//return colTex; // TODO: need to draw in 2 passes ? (mesh without contour + overlay?) - NO! just make it bg col -> on 2nd thought, maybe DOES look better if we blend decals over final image; also OUTSIDE contour?
		// single pass does not look good with bg color, need solid contour color
		//return vec4(mix(bgColor, colTex.rgb, colTex.a),1);
		color = vec4(mix(color.rgb, colTex.rgb, colTex.a),1);
	}

	return color;
}


void main() {
	ivec2 myCell = ivec2(floor(gl_FragCoord.xy));
	vec4  posVS_q  = texelFetch(texPositionVS, myCell, 0);
	vec3  posVS    = posVS_q.xyz;
	vec4  nrm_U  = texelFetch(texNormalVS,   myCell, 0);
	vec3  nrmVS  = nrm_U.xyz;
	float uParam = nrm_U.w;
	float uNormed = uParam * u_scale;
	vec4  v_flg  = texelFetch(texParamV,     myCell, 0);
	float vParam = v_flg.x;
	//bool  background = length(nrmVS) <= 0;
	float qParam = posVS_q.w; // v-param for contours
	bool  background = (qParam < 0.0);
	bool  contour = v_flg.w > 0 && hasContours;


	// NOTE!!! isotropic-trick works only with ortho projection!!
	// make v isotropic by estimating the real-world-distance (actually better: screen size distance) between 2 pixels
	// TODO: what if neighbour is background? -> not necessary to do an extensive check, this is not noticable!
	#if 0
		float dist = 0;
		dist += distance(texelFetch(texPositionVS, myCell + ivec2(-1,  0), 0).xy, posVS.xy);
		dist += distance(texelFetch(texPositionVS, myCell + ivec2( 1,  0), 0).xy, posVS.xy);
		dist += distance(texelFetch(texPositionVS, myCell + ivec2( 0, -1), 0).xy, posVS.xy);
		dist += distance(texelFetch(texPositionVS, myCell + ivec2( 0,  1), 0).xy, posVS.xy);
		dist *= 0.25; // dist ~ [0,0.01] range, (JFA) v ~= vessel radius in pixels 
	#else
		float dist = distance(texelFetch(texPositionVS, myCell + ivec2( 1, 0 ), 0).xy, posVS.xy);
	#endif
	float vScaled = vParam * dist * additional_v_scale;
	float vNormed = vScaled * u_scale; // note that this is NOT in [0,1] range!
	
	if (background) { fragColor = vec4(bgColor,1); return; }
	if (contour)    { fragColor = getContourColor(vec2(uNormed,qParam)); return; }

	// NOTE: background && contour cannot happen anymore, can it?




	vec3 color = calc_base_color(uNormed)
	           * calc_isolines_factor(uNormed, vNormed)	// isolines
		       * calc_hatching_factor(uNormed, vScaled)	// hatching
		       ;

	// add decals
	if (decals.enabled && !decals.onContour) {
		float u = uNormed * 20;
		float v = vScaled;
		float t = decals.animate ? time : 1.0;

		vec2 texUV = vec2(u,v) * decals.scale.xy + decals.offset * vec2(t,1);
		if (decals.clampV)  texUV.y = clamp(texUV.y, 0, 1);
		if (decals.invertV) texUV.y = -texUV.y;

		vec4 colTex = texture(texDecal, texUV);
		if (decals.clampV && (abs(texUV.y) >= 1.0 || abs(texUV.y) <= 0.0)) colTex = vec4(0); // mipmapping will leave artefacts otherwise
		color = mix(color, colTex.rgb, colTex.a);
	}


	// blinn-phong lighting
	float illum = 1.0;
	if (use_blinn_phong) {
		vec3 vL = -dirLight_direction_VS;
		vec3 vV = normalize(-posVS);
		vec3 vH = normalize(vL + vV);
		vec3 vN = normalize(nrmVS);
		float diff = max(dot(vL, vN), 0);
		float spec = pow(max(dot(vH, vN), 0.0), mat_shiny);
		illum = ambientLight + diff * mat_diffuse + spec * mat_specular;
	}

	fragColor = vec4(color * illum, 1);

	//fragColor = vec4(1,0,0,1);
	//fragColor = vec4(normalize(fs_in.normalWS) * 0.5 + 0.5, 1);
	//fragColor = vec4(uNormed,vNormed,vScaled,1);
}
